"""Test the sphinx.environment.adapters.toctree."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from docutils import nodes
from docutils.nodes import bullet_list, list_item, literal, reference, title

from sphinx import addnodes
from sphinx.addnodes import compact_paragraph, only
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.environment.adapters.toctree import document_toc, global_toctree_for_doc
from sphinx.testing.util import assert_node

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('xml', testroot='toctree')
@pytest.mark.test_params(shared_result='test_environment_toctree_basic')
def test_process_doc(app):
    app.build()
    # tocs
    toctree = app.env.tocs['index']
    assert_node(
        toctree,
        [
            bullet_list,
            (
                [
                    list_item,  # [0]
                    (
                        compact_paragraph,  # [0][0]
                        [
                            bullet_list,  # [0][1]
                            (
                                addnodes.toctree,  # [0][1][0]
                                only,  # [0][1][1]
                                list_item,  # [0][1][2]
                            ),
                        ],
                    ),
                ],
                [
                    list_item,  # [1]
                    (
                        compact_paragraph,  # [1][0]
                        [
                            bullet_list,  # [1][1]
                            (
                                addnodes.toctree,  # [1][1][0]
                                addnodes.toctree,  # [1][1][1]
                            ),
                        ],
                    ),
                ],
                list_item,  # [2]
            ),
        ],
    )

    assert_node(
        toctree[0][0],
        [compact_paragraph, reference, 'Welcome to Sphinx Tests’s documentation!'],
    )
    assert_node(toctree[0][0][0], reference, anchorname='')
    assert_node(
        toctree[0][1][0],
        addnodes.toctree,
        caption='Table of Contents',
        glob=False,
        hidden=False,
        titlesonly=False,
        maxdepth=2,
        numbered=999,
        entries=[
            (None, 'foo'),
            (None, 'bar'),
            (None, 'https://sphinx-doc.org/'),
            (None, 'self'),
        ],
        includefiles=['foo', 'bar'],
    )

    # only branch
    assert_node(toctree[0][1][1], addnodes.only, expr='html')
    assert_node(
        toctree[0][1][1],
        [
            only,
            list_item,
            (
                [compact_paragraph, reference, 'Section for HTML'],
                [bullet_list, addnodes.toctree],
            ),
        ],
    )
    assert_node(toctree[0][1][1][0][0][0], reference, anchorname='#section-for-html')
    assert_node(
        toctree[0][1][1][0][1][0],
        addnodes.toctree,
        caption=None,
        glob=False,
        hidden=False,
        entries=[(None, 'baz')],
        includefiles=['baz'],
        titlesonly=False,
        maxdepth=-1,
        numbered=0,
    )
    assert_node(
        toctree[0][1][2],
        (
            [compact_paragraph, reference, 'subsection'],
            [bullet_list, list_item, compact_paragraph, reference, 'subsubsection'],
        ),
    )

    assert_node(
        toctree[1][0],
        [
            compact_paragraph,
            reference,
            'Test for combination of ‘globaltoc.html’ and hidden toctree',
        ],
    )
    assert_node(
        toctree[1][0][0],
        reference,
        anchorname='#test-for-combination-of-globaltoc-html-and-hidden-toctree',
    )
    assert_node(
        toctree[1][1][0],
        addnodes.toctree,
        caption=None,
        entries=[],
        glob=False,
        hidden=False,
        titlesonly=False,
        maxdepth=-1,
        numbered=0,
    )
    assert_node(
        toctree[1][1][1],
        addnodes.toctree,
        caption=None,
        glob=False,
        hidden=True,
        titlesonly=False,
        maxdepth=-1,
        numbered=0,
        entries=[
            ('Latest reference', 'https://sphinx-doc.org/latest/'),
            ('Python', 'https://python.org/'),
        ],
    )

    assert_node(toctree[2][0], [compact_paragraph, reference, 'Indices and tables'])

    # other collections
    assert app.env.toc_num_entries['index'] == 6
    assert app.env.toctree_includes['index'] == ['foo', 'bar', 'baz']
    assert app.env.files_to_rebuild['foo'] == {'index'}
    assert app.env.files_to_rebuild['bar'] == {'index'}
    assert app.env.files_to_rebuild['baz'] == {'index'}
    assert app.env.glob_toctrees == set()
    assert app.env.numbered_toctrees == {'index'}

    # qux has no section title
    assert len(app.env.tocs['qux']) == 0
    assert_node(app.env.tocs['qux'], nodes.bullet_list)
    assert app.env.toc_num_entries['qux'] == 0
    assert 'qux' not in app.env.toctree_includes


@pytest.mark.sphinx('dummy', testroot='toctree-glob')
def test_glob(app):
    includefiles = [
        'foo',
        'bar/index',
        'bar/bar_1',
        'bar/bar_2',
        'bar/bar_3',
        'baz',
        'qux/index',
    ]

    app.build()

    # tocs
    toctree = app.env.tocs['index']
    assert_node(
        toctree,
        [
            bullet_list,
            list_item,
            (
                compact_paragraph,  # [0][0]
                [
                    bullet_list,
                    (
                        list_item,  # [0][1][0]
                        list_item,  # [0][1][1]
                    ),
                ],
            ),
        ],
    )

    assert_node(toctree[0][0], [compact_paragraph, reference, 'test-toctree-glob'])
    assert_node(
        toctree[0][1][0],
        [
            list_item,
            (
                [compact_paragraph, reference, 'normal order'],
                [bullet_list, addnodes.toctree],  # [0][1][0][1][0]
            ),
        ],
    )
    assert_node(
        toctree[0][1][0][1][0],
        addnodes.toctree,
        caption=None,
        glob=True,
        hidden=False,
        titlesonly=False,
        maxdepth=-1,
        numbered=0,
        includefiles=includefiles,
        entries=[
            (None, 'foo'),
            (None, 'bar/index'),
            (None, 'bar/bar_1'),
            (None, 'bar/bar_2'),
            (None, 'bar/bar_3'),
            (None, 'baz'),
            (None, 'qux/index'),
            ('hyperref', 'https://sphinx-doc.org/?q=sphinx'),
        ],
    )
    assert_node(
        toctree[0][1][1],
        [
            list_item,
            (
                [compact_paragraph, reference, 'reversed order'],
                [bullet_list, addnodes.toctree],  # [0][1][1][1][0]
            ),
        ],
    )
    assert_node(
        toctree[0][1][1][1][0],
        addnodes.toctree,
        caption=None,
        glob=True,
        hidden=False,
        titlesonly=False,
        maxdepth=-1,
        numbered=0,
        includefiles=list(reversed(includefiles)),
        entries=[
            (None, 'qux/index'),
            (None, 'baz'),
            (None, 'bar/bar_3'),
            (None, 'bar/bar_2'),
            (None, 'bar/bar_1'),
            (None, 'bar/index'),
            (None, 'foo'),
        ],
    )

    # other collections
    assert app.env.toc_num_entries['index'] == 3
    assert app.env.toctree_includes['index'] == includefiles + list(
        reversed(includefiles)
    )
    for file in includefiles:
        assert 'index' in app.env.files_to_rebuild[file]
    assert 'index' in app.env.glob_toctrees
    assert app.env.numbered_toctrees == set()


@pytest.mark.sphinx('dummy', testroot='toctree-domain-objects')
def test_domain_objects(app):
    app.build()

    assert app.env.toc_num_entries['index'] == 0
    assert app.env.toc_num_entries['domains'] == 9
    assert app.env.toctree_includes['index'] == ['domains', 'document_scoping']
    assert 'index' in app.env.files_to_rebuild['domains']
    assert app.env.glob_toctrees == set()
    assert app.env.numbered_toctrees == {'index'}

    # tocs
    toctree = app.env.tocs['domains']
    assert_node(
        toctree,
        [
            bullet_list,
            list_item,
            (
                compact_paragraph,  # [0][0]
                [
                    bullet_list,  # [0][1]
                    (
                        list_item,  # [0][1][0]
                        [
                            list_item,  # [0][1][1]
                            (
                                compact_paragraph,  # [0][1][1][0]
                                [
                                    bullet_list,  # [0][1][1][1]
                                    (
                                        list_item,  # [0][1][1][1][0]
                                        list_item,
                                        list_item,
                                        list_item,  # [0][1][1][1][3]
                                    ),
                                ],
                            ),
                        ],
                        list_item,
                        list_item,  # [0][1][1]
                    ),
                ],
            ),
        ],
    )

    assert_node(toctree[0][0], [compact_paragraph, reference, 'test-domain-objects'])

    assert_node(
        toctree[0][1][0],
        [list_item, ([compact_paragraph, reference, literal, 'world()'])],
    )

    assert_node(
        toctree[0][1][1][1][3],
        [
            list_item,
            ([compact_paragraph, reference, literal, 'HelloWorldPrinter.print()']),
        ],
    )


@pytest.mark.sphinx('dummy', testroot='toctree-domain-objects')
def test_domain_objects_document_scoping(app: SphinxTestApp) -> None:
    app.build()

    # tocs
    toctree = app.env.tocs['document_scoping']
    assert_node(
        toctree,
        [
            bullet_list,
            list_item,
            (
                compact_paragraph,  # [0][0]
                [
                    bullet_list,  # [0][1]
                    (
                        [
                            list_item,  # [0][1][0]
                            compact_paragraph,
                            reference,
                            literal,
                            'ClassLevel1a',
                        ],
                        [
                            list_item,  # [0][1][1]
                            (
                                [
                                    compact_paragraph,  # [0][1][1][0]
                                    reference,
                                    literal,
                                    'ClassLevel1b',
                                ],
                                [
                                    bullet_list,
                                    list_item,  # [0][1][1][1][0]
                                    compact_paragraph,
                                    reference,
                                    literal,
                                    'ClassLevel1b.f()',
                                ],
                            ),
                        ],
                        [
                            list_item,  # [0][1][2]
                            compact_paragraph,
                            reference,
                            literal,
                            'ClassLevel1a.g()',
                        ],
                        [
                            list_item,  # [0][1][3]
                            compact_paragraph,
                            reference,
                            literal,
                            'ClassLevel1b.g()',
                        ],
                        [
                            list_item,  # [0][1][4]
                            (
                                [
                                    compact_paragraph,  # [0][1][4][0]
                                    reference,
                                    'Level 2',
                                ],
                                [
                                    bullet_list,  # [0][1][4][1]
                                    (
                                        [
                                            list_item,  # [0][1][4][1][0]
                                            compact_paragraph,
                                            reference,
                                            literal,
                                            'ClassLevel2a',
                                        ],
                                        [
                                            list_item,  # [0][1][4][1][1]
                                            (
                                                [
                                                    compact_paragraph,  # [0][1][4][1][1][0]
                                                    reference,
                                                    literal,
                                                    'ClassLevel2b',
                                                ],
                                                [
                                                    bullet_list,
                                                    list_item,  # [0][1][4][1][1][1][0]
                                                    compact_paragraph,
                                                    reference,
                                                    literal,
                                                    'ClassLevel2b.f()',
                                                ],
                                            ),
                                        ],
                                        [
                                            list_item,  # [0][1][4][1][2]
                                            compact_paragraph,
                                            reference,
                                            literal,
                                            'ClassLevel2a.g()',
                                        ],
                                        [
                                            list_item,  # [0][1][4][1][3]
                                            compact_paragraph,
                                            reference,
                                            literal,
                                            'ClassLevel2b.g()',
                                        ],
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )


@pytest.mark.sphinx('xml', testroot='toctree')
@pytest.mark.test_params(shared_result='test_environment_toctree_basic')
def test_document_toc(app):
    app.build()
    toctree = document_toc(app.env, 'index', app.tags)

    assert_node(
        toctree,
        [
            bullet_list,
            (
                [
                    list_item,
                    (
                        compact_paragraph,  # [0][0]
                        [
                            bullet_list,
                            (
                                addnodes.toctree,  # [0][1][0]
                                list_item,  # [0][1][1]
                            ),
                        ],
                    ),
                ],
                [
                    list_item,
                    (
                        compact_paragraph,  # [1][0]
                        [bullet_list, (addnodes.toctree, addnodes.toctree)],
                    ),
                ],
                [list_item, compact_paragraph],  # [2][0]
            ),
        ],
    )
    assert_node(
        toctree[0][0],
        [compact_paragraph, reference, 'Welcome to Sphinx Tests’s documentation!'],
    )
    assert_node(
        toctree[0][1][1],
        (
            [compact_paragraph, reference, 'subsection'],
            [bullet_list, list_item, compact_paragraph, reference, 'subsubsection'],
        ),
    )
    assert_node(
        toctree[1][0],
        [
            compact_paragraph,
            reference,
            'Test for combination of ‘globaltoc.html’ and hidden toctree',
        ],
    )
    assert_node(toctree[2][0], [compact_paragraph, reference, 'Indices and tables'])


@pytest.mark.sphinx('xml', testroot='toctree')
@pytest.mark.test_params(shared_result='test_environment_toctree_basic')
def test_document_toc_only(app):
    app.build()
    StandaloneHTMLBuilder(app, app.env)  # adds format/builder tags
    toctree = document_toc(app.env, 'index', app.tags)

    assert_node(
        toctree,
        [
            bullet_list,
            (
                [
                    list_item,
                    (
                        compact_paragraph,  # [0][0]
                        [
                            bullet_list,
                            (
                                addnodes.toctree,  # [0][1][0]
                                list_item,  # [0][1][1]
                                list_item,  # [0][1][2]
                            ),
                        ],
                    ),
                ],
                [
                    list_item,
                    (
                        compact_paragraph,  # [1][0]
                        [bullet_list, (addnodes.toctree, addnodes.toctree)],
                    ),
                ],
                [list_item, compact_paragraph],  # [2][0]
            ),
        ],
    )
    assert_node(
        toctree[0][0],
        [compact_paragraph, reference, 'Welcome to Sphinx Tests’s documentation!'],
    )
    assert_node(
        toctree[0][1][1],
        (
            [compact_paragraph, reference, 'Section for HTML'],
            [bullet_list, addnodes.toctree],
        ),
    )
    assert_node(
        toctree[0][1][2],
        (
            [compact_paragraph, reference, 'subsection'],
            [bullet_list, list_item, compact_paragraph, reference, 'subsubsection'],
        ),
    )
    assert_node(
        toctree[1][0],
        [
            compact_paragraph,
            reference,
            'Test for combination of ‘globaltoc.html’ and hidden toctree',
        ],
    )
    assert_node(toctree[2][0], [compact_paragraph, reference, 'Indices and tables'])


@pytest.mark.sphinx('xml', testroot='toctree')
@pytest.mark.test_params(shared_result='test_environment_toctree_basic')
def test_document_toc_tocdepth(app):
    app.build()
    toctree = document_toc(app.env, 'tocdepth', app.tags)

    assert_node(
        toctree,
        [
            bullet_list,
            list_item,
            (
                compact_paragraph,  # [0][0]
                bullet_list,  # [0][1]
            ),
        ],
    )
    assert_node(toctree[0][0], [compact_paragraph, reference, 'level 1'])
    assert_node(
        toctree[0][1], [bullet_list, list_item, compact_paragraph, reference, 'level 2']
    )


@pytest.mark.sphinx('xml', testroot='toctree')
@pytest.mark.test_params(shared_result='test_environment_toctree_basic')
def test_global_toctree_for_doc(app):
    app.build()
    toctree = global_toctree_for_doc(app.env, 'index', app.builder, collapse=False)
    assert_node(
        toctree,
        [
            compact_paragraph,
            ([title, 'Table of Contents'], bullet_list, bullet_list, bullet_list),
        ],
    )

    assert_node(
        toctree[1],
        (
            [list_item, ([compact_paragraph, reference, 'foo'], bullet_list)],
            [list_item, compact_paragraph, reference, 'bar'],
            [list_item, compact_paragraph, reference, 'https://sphinx-doc.org/'],
            [
                list_item,
                compact_paragraph,
                reference,
                'Welcome to Sphinx Tests’s documentation!',
            ],
        ),
    )
    assert_node(
        toctree[1][0][1],
        (
            [list_item, compact_paragraph, reference, 'quux'],
            [list_item, compact_paragraph, reference, 'foo.1'],
            [list_item, compact_paragraph, reference, 'foo.2'],
        ),
    )

    assert_node(toctree[1][0][0][0], reference, refuri='foo', secnumber=[1])
    assert_node(toctree[1][0][1][0][0][0], reference, refuri='quux', secnumber=[1, 1])
    assert_node(
        toctree[1][0][1][1][0][0], reference, refuri='foo#foo-1', secnumber=[1, 2]
    )
    assert_node(
        toctree[1][0][1][2][0][0], reference, refuri='foo#foo-2', secnumber=[1, 3]
    )
    assert_node(toctree[1][1][0][0], reference, refuri='bar', secnumber=[2])
    assert_node(toctree[1][2][0][0], reference, refuri='https://sphinx-doc.org/')
    assert_node(toctree[1][3][0][0], reference, refuri='')

    assert_node(
        toctree[2], [bullet_list, list_item, compact_paragraph, reference, 'baz']
    )
    assert_node(
        toctree[3],
        (
            [list_item, compact_paragraph, reference, 'Latest reference'],
            [list_item, compact_paragraph, reference, 'Python'],
        ),
    )
    assert_node(toctree[3][0][0][0], reference, refuri='https://sphinx-doc.org/latest/')
    assert_node(toctree[3][1][0][0], reference, refuri='https://python.org/')


@pytest.mark.sphinx('xml', testroot='toctree')
@pytest.mark.test_params(shared_result='test_environment_toctree_basic')
def test_global_toctree_for_doc_collapse(app):
    app.build()
    toctree = global_toctree_for_doc(app.env, 'index', app.builder, collapse=True)
    assert_node(
        toctree,
        [
            compact_paragraph,
            ([title, 'Table of Contents'], bullet_list, bullet_list, bullet_list),
        ],
    )

    assert_node(
        toctree[1],
        (
            [list_item, compact_paragraph, reference, 'foo'],
            [list_item, compact_paragraph, reference, 'bar'],
            [list_item, compact_paragraph, reference, 'https://sphinx-doc.org/'],
            [
                list_item,
                compact_paragraph,
                reference,
                'Welcome to Sphinx Tests’s documentation!',
            ],
        ),
    )
    assert_node(toctree[1][0][0][0], reference, refuri='foo', secnumber=[1])
    assert_node(toctree[1][1][0][0], reference, refuri='bar', secnumber=[2])
    assert_node(toctree[1][2][0][0], reference, refuri='https://sphinx-doc.org/')
    assert_node(toctree[1][3][0][0], reference, refuri='')

    assert_node(
        toctree[2], [bullet_list, list_item, compact_paragraph, reference, 'baz']
    )
    assert_node(
        toctree[3],
        (
            [list_item, compact_paragraph, reference, 'Latest reference'],
            [list_item, compact_paragraph, reference, 'Python'],
        ),
    )
    assert_node(toctree[3][0][0][0], reference, refuri='https://sphinx-doc.org/latest/')
    assert_node(toctree[3][1][0][0], reference, refuri='https://python.org/')


@pytest.mark.sphinx('xml', testroot='toctree')
@pytest.mark.test_params(shared_result='test_environment_toctree_basic')
def test_global_toctree_for_doc_maxdepth(app):
    app.build()
    toctree = global_toctree_for_doc(
        app.env, 'index', app.builder, collapse=False, maxdepth=3
    )
    assert_node(
        toctree,
        [
            compact_paragraph,
            (
                [title, 'Table of Contents'],
                bullet_list,
                bullet_list,
                bullet_list,
            ),
        ],
    )

    assert_node(
        toctree[1],
        (
            [
                list_item,
                (
                    [compact_paragraph, reference, 'foo'],
                    bullet_list,
                ),
            ],
            [list_item, compact_paragraph, reference, 'bar'],
            [list_item, compact_paragraph, reference, 'https://sphinx-doc.org/'],
            [
                list_item,
                compact_paragraph,
                reference,
                'Welcome to Sphinx Tests’s documentation!',
            ],
        ),
    )
    assert_node(
        toctree[1][0][1],
        (
            [list_item, compact_paragraph, reference, 'quux'],
            [
                list_item,
                (
                    [compact_paragraph, reference, 'foo.1'],
                    bullet_list,
                ),
            ],
            [list_item, compact_paragraph, reference, 'foo.2'],
        ),
    )
    assert_node(
        toctree[1][0][1][1][1],
        [bullet_list, list_item, compact_paragraph, reference, 'foo.1-1'],
    )

    assert_node(toctree[1][0][0][0], reference, refuri='foo', secnumber=[1])
    assert_node(toctree[1][0][1][0][0][0], reference, refuri='quux', secnumber=[1, 1])
    assert_node(
        toctree[1][0][1][1][0][0], reference, refuri='foo#foo-1', secnumber=[1, 2]
    )
    assert_node(
        toctree[1][0][1][1][1][0][0][0],
        reference,
        refuri='foo#foo-1-1',
        secnumber=[1, 2, 1],
    )
    assert_node(
        toctree[1][0][1][2][0][0], reference, refuri='foo#foo-2', secnumber=[1, 3]
    )
    assert_node(toctree[1][1][0][0], reference, refuri='bar', secnumber=[2])
    assert_node(toctree[1][2][0][0], reference, refuri='https://sphinx-doc.org/')
    assert_node(toctree[1][3][0][0], reference, refuri='')

    assert_node(
        toctree[2], [bullet_list, list_item, compact_paragraph, reference, 'baz']
    )
    assert_node(
        toctree[3],
        (
            [list_item, compact_paragraph, reference, 'Latest reference'],
            [list_item, compact_paragraph, reference, 'Python'],
        ),
    )
    assert_node(toctree[3][0][0][0], reference, refuri='https://sphinx-doc.org/latest/')
    assert_node(toctree[3][1][0][0], reference, refuri='https://python.org/')


@pytest.mark.sphinx('xml', testroot='toctree')
@pytest.mark.test_params(shared_result='test_environment_toctree_basic')
def test_global_toctree_for_doc_includehidden(app):
    app.build()
    toctree = global_toctree_for_doc(
        app.env, 'index', app.builder, collapse=False, includehidden=False
    )
    assert_node(
        toctree,
        [
            compact_paragraph,
            (
                [title, 'Table of Contents'],
                bullet_list,
                bullet_list,
            ),
        ],
    )

    assert_node(
        toctree[1],
        (
            [
                list_item,
                (
                    [compact_paragraph, reference, 'foo'],
                    bullet_list,
                ),
            ],
            [list_item, compact_paragraph, reference, 'bar'],
            [list_item, compact_paragraph, reference, 'https://sphinx-doc.org/'],
            [
                list_item,
                compact_paragraph,
                reference,
                'Welcome to Sphinx Tests’s documentation!',
            ],
        ),
    )
    assert_node(
        toctree[1][0][1],
        (
            [list_item, compact_paragraph, reference, 'quux'],
            [list_item, compact_paragraph, reference, 'foo.1'],
            [list_item, compact_paragraph, reference, 'foo.2'],
        ),
    )

    assert_node(toctree[1][0][0][0], reference, refuri='foo', secnumber=[1])
    assert_node(toctree[1][0][1][0][0][0], reference, refuri='quux', secnumber=[1, 1])
    assert_node(
        toctree[1][0][1][1][0][0], reference, refuri='foo#foo-1', secnumber=[1, 2]
    )
    assert_node(
        toctree[1][0][1][2][0][0], reference, refuri='foo#foo-2', secnumber=[1, 3]
    )
    assert_node(toctree[1][1][0][0], reference, refuri='bar', secnumber=[2])
    assert_node(toctree[1][2][0][0], reference, refuri='https://sphinx-doc.org/')

    assert_node(
        toctree[2], [bullet_list, list_item, compact_paragraph, reference, 'baz']
    )


@pytest.mark.sphinx('xml', testroot='toctree-index')
def test_toctree_index(app):
    app.build()
    toctree = app.env.tocs['index']
    assert_node(
        toctree,
        [
            bullet_list,
            ([
                list_item,
                (
                    compact_paragraph,  # [0][0]
                    [
                        bullet_list,
                        (
                            addnodes.toctree,  # [0][1][0]
                            addnodes.toctree,  # [0][1][1]
                        ),
                    ],
                ),
            ]),
        ],
    )
    assert_node(
        toctree[0][1][1],
        addnodes.toctree,
        caption='Indices',
        glob=False,
        hidden=False,
        titlesonly=False,
        maxdepth=-1,
        numbered=0,
        entries=[(None, 'genindex'), (None, 'modindex'), (None, 'search')],
    )



"""Test smart quotes."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sphinx.testing.util import etree_parse

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx(
    'html',
    testroot='smartquotes',
    freshenv=True,
)
def test_basic(app: SphinxTestApp) -> None:
    app.build()

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<p>– “Sphinx” is a tool that makes it easy …</p>' in content


@pytest.mark.sphinx(
    'html',
    testroot='smartquotes',
    freshenv=True,
)
def test_literals(app: SphinxTestApp) -> None:
    app.build()

    etree = etree_parse(app.outdir / 'literals.html')
    for code_element in etree.iter('code'):
        code_text = ''.join(code_element.itertext())

        if code_text.startswith('code role'):
            assert "'quotes'" in code_text
        elif code_text.startswith('{'):
            assert code_text == "{'code': 'role', 'with': 'quotes'}"
        elif code_text.startswith('literal'):
            assert code_text == "literal with 'quotes'"


@pytest.mark.sphinx(
    'text',
    testroot='smartquotes',
    freshenv=True,
)
def test_text_builder(app: SphinxTestApp) -> None:
    app.build()

    content = (app.outdir / 'index.txt').read_text(encoding='utf8')
    assert '-- "Sphinx" is a tool that makes it easy ...' in content


@pytest.mark.sphinx(
    'man',
    testroot='smartquotes',
    freshenv=True,
)
def test_man_builder(app: SphinxTestApp) -> None:
    app.build()

    content = (app.outdir / 'projectnamenotset.1').read_text(encoding='utf8')
    assert r'\-\- \(dqSphinx\(dq is a tool that makes it easy ...' in content


@pytest.mark.sphinx(
    'latex',
    testroot='smartquotes',
    freshenv=True,
)
def test_latex_builder(app: SphinxTestApp) -> None:
    app.build()

    content = (app.outdir / 'projectnamenotset.tex').read_text(encoding='utf8')
    assert '\\textendash{} “Sphinx” is a tool that makes it easy …' in content


@pytest.mark.sphinx(
    'html',
    testroot='smartquotes',
    freshenv=True,
    confoverrides={'language': 'ja'},
)
def test_ja_html_builder(app: SphinxTestApp) -> None:
    app.build()

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<p>-- &quot;Sphinx&quot; is a tool that makes it easy ...</p>' in content


@pytest.mark.sphinx(
    'html',
    testroot='smartquotes',
    freshenv=True,
    confoverrides={'language': 'zh_CN'},
)
def test_zh_cn_html_builder(app: SphinxTestApp) -> None:
    app.build()

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<p>-- &quot;Sphinx&quot; is a tool that makes it easy ...</p>' in content


@pytest.mark.sphinx(
    'html',
    testroot='smartquotes',
    freshenv=True,
    confoverrides={'language': 'zh_TW'},
)
def test_zh_tw_html_builder(app: SphinxTestApp) -> None:
    app.build()

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<p>-- &quot;Sphinx&quot; is a tool that makes it easy ...</p>' in content


@pytest.mark.sphinx(
    'html',
    testroot='smartquotes',
    freshenv=True,
    confoverrides={'smartquotes': False},
)
def test_smartquotes_disabled(app: SphinxTestApp) -> None:
    app.build()

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<p>-- &quot;Sphinx&quot; is a tool that makes it easy ...</p>' in content


@pytest.mark.sphinx(
    'html',
    testroot='smartquotes',
    freshenv=True,
    confoverrides={'smartquotes_action': 'q'},
)
def test_smartquotes_action(app: SphinxTestApp) -> None:
    app.build()

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<p>-- “Sphinx” is a tool that makes it easy ...</p>' in content


@pytest.mark.sphinx(
    'html',
    testroot='smartquotes',
    freshenv=True,
    confoverrides={'language': 'ja', 'smartquotes_excludes': {}},
)
def test_smartquotes_excludes_language(app: SphinxTestApp) -> None:
    app.build()

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<p>– 「Sphinx」 is a tool that makes it easy …</p>' in content


@pytest.mark.sphinx(
    'man',
    testroot='smartquotes',
    freshenv=True,
    confoverrides={'smartquotes_excludes': {}},
)
def test_smartquotes_excludes_builders(app: SphinxTestApp) -> None:
    app.build()

    content = (app.outdir / 'projectnamenotset.1').read_text(encoding='utf8')
    assert '– “Sphinx” is a tool that makes it easy …' in content



"""Test the HTML builder and check output against XPath."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

import pytest
from docutils import nodes

from tests.test_builders.xpath_util import check_xpath

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from typing import Literal
    from xml.etree.ElementTree import Element

    from sphinx.testing.util import SphinxTestApp


def tail_check(check: str) -> Callable[[Iterable[Element]], Literal[True]]:
    rex = re.compile(check)

    def checker(nodes: Iterable[Element]) -> Literal[True]:
        for node in nodes:
            if node.tail and rex.search(node.tail):
                return True
        msg = f'{check!r} not found in tail of any nodes {nodes}'
        raise AssertionError(msg)

    return checker


@pytest.mark.parametrize(
    ('fname', 'path', 'check'),
    [
        ('images.html', ".//img[@src='_images/img.png']", ''),
        ('images.html', ".//img[@src='_images/img1.png']", ''),
        ('images.html', ".//img[@src='_images/simg.png']", ''),
        ('images.html', ".//img[@src='_images/svgimg.svg']", ''),
        ('images.html', ".//a[@href='_sources/images.txt']", ''),
        # Check svg options
        ('images.html', ".//img[@src='_images/svgimg.svg'][@style='width: 2cm;']", ''),
        ('images.html', ".//img[@src='_images/svgimg.svg'][@style='height: 2cm;']", ''),
        ('subdir/images.html', ".//img[@src='../_images/img1.png']", ''),
        ('subdir/images.html', ".//img[@src='../_images/rimg.png']", ''),
        ('subdir/includes.html', ".//a[@class='reference download internal']", ''),
        ('subdir/includes.html', ".//img[@src='../_images/img.png']", ''),
        ('subdir/includes.html', './/p', 'This is an include file.'),
        ('subdir/includes.html', './/pre/span', 'line 1'),
        ('subdir/includes.html', './/pre/span', 'line 2'),
        ('includes.html', './/pre', 'Max Strauß'),
        ('includes.html', ".//a[@class='reference download internal']", ''),
        ('includes.html', './/pre/span', '"quotes"'),
        ('includes.html', './/pre/span', "'included'"),
        ('includes.html', ".//pre/span[@class='s2']", 'üöä'),
        (
            'includes.html',
            ".//div[@class='inc-pyobj1 highlight-text notranslate']//pre",
            r'^class Foo:\n    pass\n\s*$',
        ),
        (
            'includes.html',
            ".//div[@class='inc-pyobj2 highlight-text notranslate']//pre",
            r'^    def baz\(\):\n        pass\n\s*$',
        ),
        (
            'includes.html',
            ".//div[@class='inc-lines highlight-text notranslate']//pre",
            r'^class Foo:\n    pass\nclass Bar:\n$',
        ),
        (
            'includes.html',
            ".//div[@class='inc-startend highlight-text notranslate']//pre",
            '^foo = "Including Unicode characters: üöä"\\n$',
        ),
        (
            'includes.html',
            ".//div[@class='inc-preappend highlight-text notranslate']//pre",
            r'(?m)^START CODE$',
        ),
        (
            'includes.html',
            ".//div[@class='inc-pyobj-dedent highlight-python notranslate']//span",
            r'def',
        ),
        (
            'includes.html',
            ".//div[@class='inc-tab3 highlight-text notranslate']//pre",
            r'-| |-',
        ),
        (
            'includes.html',
            ".//div[@class='inc-tab8 highlight-python notranslate']//pre/span",
            r'-|      |-',
        ),
        ('autodoc.html', ".//dl[@class='py class']/dt[@id='autodoc_target.Class']", ''),
        (
            'autodoc.html',
            ".//dl[@class='py function']/dt[@id='autodoc_target.function']/em/span/span",
            r'\*\*',
        ),
        (
            'autodoc.html',
            ".//dl[@class='py function']/dt[@id='autodoc_target.function']/em/span/span",
            r'kwds',
        ),
        ('autodoc.html', './/dd/p', r'Return spam\.'),
        ('extapi.html', './/strong', 'from class: Bar'),
        ('markup.html', './/title', 'set by title directive'),
        ('markup.html', './/p/em', 'Section author: Georg Brandl'),
        ('markup.html', './/p/em', 'Module author: Georg Brandl'),
        # created by the meta directive
        ('markup.html', ".//meta[@name='author'][@content='Me']", ''),
        ('markup.html', ".//meta[@name='keywords'][@content='docs, sphinx']", ''),
        # a label created by ``.. _label:``
        ('markup.html', ".//div[@id='label']", ''),
        # code with standard code blocks
        ('markup.html', './/pre', '^some code$'),
        # an option list
        ('markup.html', ".//span[@class='option']", '--help'),
        # admonitions
        ('markup.html', ".//p[@class='admonition-title']", 'My Admonition'),
        ('markup.html', ".//div[@class='admonition note']/p", 'Note text.'),
        ('markup.html', ".//div[@class='admonition warning']/p", 'Warning text.'),
        # inline markup
        ('markup.html', './/li/p/strong', r'^command\\n$'),
        ('markup.html', './/li/p/strong', r'^program\\n$'),
        ('markup.html', './/li/p/em', r'^dfn\\n$'),
        ('markup.html', './/li/p/kbd', r'^kbd\\n$'),
        ('markup.html', './/li/p/span', 'File \N{TRIANGULAR BULLET} Close'),
        ('markup.html', ".//li/p/code/span[@class='pre']", '^a/$'),
        ('markup.html', ".//li/p/code/em/span[@class='pre']", '^varpart$'),
        ('markup.html', ".//li/p/code/em/span[@class='pre']", '^i$'),
        (
            'markup.html',
            ".//a[@href='https://peps.python.org/pep-0008/']"
            "[@class='pep reference external']/strong",
            'PEP 8',
        ),
        (
            'markup.html',
            ".//a[@href='https://peps.python.org/pep-0008/']"
            "[@class='pep reference external']/strong",
            'Python Enhancement Proposal #8',
        ),
        (
            'markup.html',
            ".//a[@href='https://datatracker.ietf.org/doc/html/rfc1.html']"
            "[@class='rfc reference external']/strong",
            'RFC 1',
        ),
        (
            'markup.html',
            ".//a[@href='https://datatracker.ietf.org/doc/html/rfc1.html']"
            "[@class='rfc reference external']/strong",
            'Request for Comments #1',
        ),
        (
            'markup.html',
            ".//a[@href='objects.html#envvar-HOME']"
            "[@class='reference internal']/code/span[@class='pre']",
            'HOME',
        ),
        (
            'markup.html',
            ".//a[@href='#with'][@class='reference internal']/code/span[@class='pre']",
            '^with$',
        ),
        (
            'markup.html',
            ".//a[@href='#grammar-token-try_stmt']"
            "[@class='reference internal']/code/span",
            '^statement$',
        ),
        (
            'markup.html',
            ".//a[@href='#some-label'][@class='reference internal']/span",
            '^here$',
        ),
        (
            'markup.html',
            ".//a[@href='#some-label'][@class='reference internal']/span",
            '^there$',
        ),
        (
            'markup.html',
            ".//a[@href='subdir/includes.html'][@class='reference internal']/span",
            'Including in subdir',
        ),
        (
            'markup.html',
            ".//a[@href='objects.html#cmdoption-python-c']"
            "[@class='reference internal']/code/span[@class='pre']",
            '-c',
        ),
        # abbreviations
        ('markup.html', ".//abbr[@title='abbreviation']", '^abbr$'),
        # version stuff
        (
            'markup.html',
            ".//div[@class='versionadded']/p/span",
            'Added in version 0.6: ',
        ),
        (
            'markup.html',
            ".//div[@class='versionadded']/p/span",
            tail_check('First paragraph of versionadded'),
        ),
        (
            'markup.html',
            ".//div[@class='versionchanged']/p/span",
            tail_check('First paragraph of versionchanged'),
        ),
        (
            'markup.html',
            ".//div[@class='versionchanged']/p",
            'Second paragraph of versionchanged',
        ),
        (
            'markup.html',
            ".//div[@class='versionremoved']/p/span",
            'Removed in version 0.6: ',
        ),
        # footnote reference
        ('markup.html', ".//a[@class='footnote-reference brackets']", r'1'),
        # created by reference lookup
        ('markup.html', ".//a[@href='index.html#ref1']", ''),
        # ``seealso`` directive
        ('markup.html', ".//div/p[@class='admonition-title']", 'See also'),
        # a ``hlist`` directive
        ('markup.html', ".//table[@class='hlist']/tr/td/ul/li/p", '^This$'),
        # a ``centered`` directive
        ('markup.html', ".//p[@class='centered']/strong", 'LICENSE'),
        # a glossary
        ('markup.html', ".//dl/dt[@id='term-boson']", 'boson'),
        ('markup.html', ".//dl/dt[@id='term-boson']/a", '¶'),
        # a production list
        ('markup.html', './/pre/strong', 'try_stmt'),
        (
            'markup.html',
            ".//pre/a[@href='#grammar-token-try1_stmt']/code/span",
            'try1_stmt',
        ),
        # tests for ``only`` directive
        ('markup.html', './/p', 'A global substitution!'),
        ('markup.html', './/p', 'In HTML.'),
        ('markup.html', './/p', 'In both.'),
        ('markup.html', './/p', 'Always present'),
        # tests for ``any`` role
        ('markup.html', ".//a[@href='#with']/span", 'headings'),
        (
            'markup.html',
            ".//a[@href='objects.html#func_without_body']/code/span",
            'objects',
        ),
        # tests for numeric labels
        (
            'markup.html',
            ".//a[@href='#id1'][@class='reference internal']/span",
            'Testing various markup',
        ),
        # tests for smartypants
        ('markup.html', './/li/p', 'Smart “quotes” in English ‘text’.'),
        ('markup.html', './/li/p', 'Smart — long and – short dashes.'),
        ('markup.html', './/li/p', 'Ellipsis…'),
        ('markup.html', ".//li/p/code/span[@class='pre']", 'foo--"bar"...'),
        ('markup.html', './/p', 'Этот «абзац» должен использовать „русские“ кавычки.'),
        ('markup.html', './/p', 'Il dit : « C’est “super” ! »'),
        ('objects.html', ".//dt[@id='mod.Cls.meth1']", ''),
        ('objects.html', ".//dt[@id='errmod.Error']", ''),
        (
            'objects.html',
            ".//dt/span[@class='sig-name descname']/span[@class='pre']",
            r'long\(parameter,',
        ),
        (
            'objects.html',
            ".//dt/span[@class='sig-name descname']/span[@class='pre']",
            r'list\)',
        ),
        (
            'objects.html',
            ".//dt/span[@class='sig-name descname']/span[@class='pre']",
            'another',
        ),
        (
            'objects.html',
            ".//dt/span[@class='sig-name descname']/span[@class='pre']",
            'one',
        ),
        ('objects.html', ".//a[@href='#mod.Cls'][@class='reference internal']", ''),
        ('objects.html', ".//dl[@class='std userdesc']", ''),
        ('objects.html', ".//dt[@id='userdesc-myobj']", ''),
        (
            'objects.html',
            ".//a[@href='#userdesc-myobj'][@class='reference internal']",
            '',
        ),
        # docfields
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#TimeInt']/em",
            'TimeInt',
        ),
        ('objects.html', ".//a[@class='reference internal'][@href='#Time']", 'Time'),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#errmod.Error']/strong",
            'Error',
        ),
        # C references
        ('objects.html', ".//span[@class='pre']", 'CFunction()'),
        ('objects.html', ".//a[@href='#c.Sphinx_DoSomething']", ''),
        ('objects.html', ".//a[@href='#c.SphinxStruct.member']", ''),
        ('objects.html', ".//a[@href='#c.SPHINX_USE_PYTHON']", ''),
        ('objects.html', ".//a[@href='#c.SphinxType']", ''),
        ('objects.html', ".//a[@href='#c.sphinx_global']", ''),
        # test global TOC created by toctree()
        (
            'objects.html',
            ".//ul[@class='current']/li[@class='toctree-l1 current']/a[@href='#']",
            'Testing object descriptions',
        ),
        (
            'objects.html',
            ".//li[@class='toctree-l1']/a[@href='markup.html']",
            'Testing various markup',
        ),
        # test unknown field names
        ('objects.html', ".//dt[@class='field-odd']", 'Field_name'),
        ('objects.html', ".//dt[@class='field-even']", 'Field_name all lower'),
        ('objects.html', ".//dt[@class='field-odd']", 'FIELD_NAME'),
        ('objects.html', ".//dt[@class='field-even']", 'FIELD_NAME ALL CAPS'),
        ('objects.html', ".//dt[@class='field-odd']", 'Field_Name'),
        ('objects.html', ".//dt[@class='field-even']", 'Field_Name All Word Caps'),
        # ('objects.html', ".//dt[@class='field-odd']", 'Field_name'), (duplicate)
        ('objects.html', ".//dt[@class='field-even']", 'Field_name First word cap'),
        ('objects.html', ".//dt[@class='field-odd']", 'FIELd_name'),
        ('objects.html', ".//dt[@class='field-even']", 'FIELd_name PARTial caps'),
        # custom sidebar
        ('objects.html', './/h4', 'Custom sidebar'),
        # docfields
        ('objects.html', ".//dd[@class='field-odd']/p/strong", '^moo$'),
        (
            'objects.html',
            ".//dd[@class='field-odd']/p/strong",
            tail_check(r'\(Moo\) .* Moo'),
        ),
        ('objects.html', ".//dd[@class='field-odd']/ul/li/p/strong", '^hour$'),
        ('objects.html', ".//dd[@class='field-odd']/ul/li/p/em", '^DuplicateType$'),
        (
            'objects.html',
            ".//dd[@class='field-odd']/ul/li/p/em",
            tail_check(r'.* Some parameter'),
        ),
        # others
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-perl-arg-p']/code/span",
            'perl',
        ),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-perl-arg-p']/code/span",
            '\\+p',
        ),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-perl-ObjC']/code/span",
            '--ObjC\\+\\+',
        ),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-perl-plugin.option']/code/span",
            '--plugin.option',
        ),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-perl-arg-create-auth-token']"
            '/code/span',
            'create-auth-token',
        ),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-perl-arg-arg']/code/span",
            'arg',
        ),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-perl-j']/code/span",
            '-j',
        ),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-hg-arg-commit']/code/span",
            'hg',
        ),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-hg-arg-commit']/code/span",
            'commit',
        ),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-git-commit-p']/code/span",
            'git',
        ),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-git-commit-p']/code/span",
            'commit',
        ),
        (
            'objects.html',
            ".//a[@class='reference internal'][@href='#cmdoption-git-commit-p']/code/span",
            '-p',
        ),
        ('index.html', ".//meta[@name='hc'][@content='hcval']", ''),
        ('index.html', ".//meta[@name='hc_co'][@content='hcval_co']", ''),
        ('index.html', ".//li[@class='toctree-l1']/a", 'Testing various markup'),
        ('index.html', ".//li[@class='toctree-l2']/a", 'Inline markup'),
        ('index.html', './/title', 'Sphinx <Tests>'),
        ('index.html', ".//div[@class='footer']", 'copyright text credits'),
        (
            'index.html',
            ".//a[@href='https://python.org/'][@class='reference external']",
            '',
        ),
        ('index.html', ".//li/p/a[@href='genindex.html']/span", 'Index'),
        ('index.html', ".//li/p/a[@href='py-modindex.html']/span", 'Module Index'),
        # custom sidebar only for contents
        ('index.html', './/h4', 'Contents sidebar'),
        # custom JavaScript
        ('index.html', ".//script[@src='file://moo.js']", ''),
        # URL in contents
        (
            'index.html',
            ".//a[@class='reference external'][@href='https://sphinx-doc.org/']",
            'https://sphinx-doc.org/',
        ),
        (
            'index.html',
            ".//a[@class='reference external'][@href='https://sphinx-doc.org/latest/']",
            'Latest reference',
        ),
        # Indirect hyperlink targets across files
        (
            'index.html',
            ".//a[@href='markup.html#some-label'][@class='reference internal']/span",
            '^indirect hyperref$',
        ),
        ('bom.html', './/title', ' File with UTF-8 BOM'),
        (
            'extensions.html',
            ".//a[@href='https://python.org/dev/']",
            'https://python.org/dev/',
        ),
        (
            'extensions.html',
            ".//a[@href='https://bugs.python.org/issue1000']",
            'issue 1000',
        ),
        (
            'extensions.html',
            ".//a[@href='https://bugs.python.org/issue1042']",
            'explicit caption',
        ),
        (
            'extensions.html',
            ".//a[@class='extlink-pyurl reference external']",
            'https://python.org/dev/',
        ),
        (
            'extensions.html',
            ".//a[@class='extlink-issue reference external']",
            'issue 1000',
        ),
        # index entries
        ('genindex.html', './/a/strong', 'Main'),
        ('genindex.html', './/a/strong', '[1]'),
        ('genindex.html', './/a/strong', 'Other'),
        ('genindex.html', './/a', 'entry'),
        ('genindex.html', './/li/a', 'double'),
        ('otherext.html', './/h1', 'Generated section'),
        ('otherext.html', ".//a[@href='_sources/otherext.foo.txt']", ''),
        ('search.html', ".//meta[@name='robots'][@content='noindex']", ''),
    ],
)
@pytest.mark.sphinx(
    'html',
    testroot='root',
    tags=['testtag'],
    confoverrides={'html_context.hckey_co': 'hcval_co'},
)
def test_html5_output(app, cached_etree_parse, fname, path, check):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check)


@pytest.mark.sphinx('html', testroot='markup-rubric')
def test_html5_rubric(app: SphinxTestApp) -> None:
    def insert_invalid_rubric_heading_level(app, doctree, docname):
        if docname != 'index':
            return
        new_node = nodes.rubric('', 'INSERTED RUBRIC')
        new_node['heading-level'] = 7
        doctree[0].append(new_node)

    app.connect('doctree-resolved', insert_invalid_rubric_heading_level)
    app.build()

    warnings = app.warning.getvalue()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<p class="rubric">This is a rubric</p>' in content
    assert '<h2 class="myclass rubric">A rubric with a heading level 2</h2>' in content

    # directive warning
    assert '"7" unknown' in warnings

    # html writer warning
    assert 'WARNING: unsupported rubric heading level: 7' in warnings
    assert '</h7>' not in content
    assert '<p class="rubric">INSERTED RUBRIC</p>' in content



from __future__ import annotations

import re
import textwrap
from typing import TYPE_CHECKING
from xml.etree.ElementTree import tostring

if TYPE_CHECKING:
    import os
    from collections.abc import Callable, Iterable, Sequence
    from xml.etree.ElementTree import Element, ElementTree


def _get_text(node: Element) -> str:
    if node.text is not None:
        # the node has only one text
        return node.text

    # the node has tags and text; gather texts just under the node
    return ''.join(n.tail or '' for n in node)


def _prettify(nodes: Iterable[Element]) -> str:
    def pformat(node: Element) -> str:
        return tostring(node, encoding='unicode', method='html')

    return ''.join(f'(i={index}) {pformat(node)}\n' for index, node in enumerate(nodes))


def check_xpath(
    etree: ElementTree,
    filename: str | os.PathLike[str],
    xpath: str,
    check: str | re.Pattern[str] | Callable[[Sequence[Element]], None] | None,
    be_found: bool = True,
    *,
    min_count: int = 1,
) -> None:
    """Check that one or more nodes satisfy a predicate.

    :param etree: The element tree.
    :param filename: The element tree source name (for errors only).
    :param xpath: An XPath expression to use.
    :param check: Optional regular expression or a predicate the nodes must validate.
    :param be_found: If false, negate the predicate.
    :param min_count: Minimum number of nodes expected to satisfy the predicate.

    * If *check* is empty (``''``), only the minimum count is checked.
    * If *check* is ``None``, no node should satisfy the XPath expression.
    """
    nodes = etree.findall(xpath)
    assert isinstance(nodes, list)

    if check is None:
        # use == to have a nice pytest diff
        assert nodes == [], f'found nodes matching xpath {xpath!r} in file {filename}'
        return

    assert len(nodes) >= min_count, (
        f'expecting at least {min_count} node(s) '
        f'to satisfy {xpath!r} in file {filename}'
    )

    if check == '':
        return

    if callable(check):
        check(nodes)
        return

    rex = re.compile(check)
    if be_found:
        if any(rex.search(_get_text(node)) for node in nodes):
            return
    else:
        if all(not rex.search(_get_text(node)) for node in nodes):
            return

    ctx = textwrap.indent(_prettify(nodes), ' ' * 2)
    msg = (
        f'{check!r} not found in any node matching {xpath!r} in file {filename}:\n{ctx}'
    )
    raise AssertionError(msg)



from __future__ import annotations

from datetime import date
from typing import NewType, TypeVar

#: T1
T1 = TypeVar('T1')

T2 = TypeVar('T2')  # A TypeVar not having doc comment

#: T3
T3 = TypeVar('T3', int, str)

#: T4
T4 = TypeVar('T4', covariant=True)  # NoQA: PLC0105

#: T5
T5 = TypeVar('T5', contravariant=True)  # NoQA: PLC0105

#: T6
T6 = NewType('T6', date)

#: T7
T7 = TypeVar('T7', bound=int)


class Class:
    #: T1
    T1 = TypeVar('T1')

    #: T6
    T6 = NewType('T6', date)



"""Test the HTML builder and check output against XPath."""

from __future__ import annotations

import pytest

from tests.test_builders.xpath_html_util import _intradocument_hyperlink_check
from tests.test_builders.xpath_util import check_xpath


@pytest.mark.parametrize(
    ('fname', 'path', 'check', 'be_found'),
    [
        # index.html
        ('index.html', ".//li[@class='toctree-l3']/a", '1.1.1. Foo A1', True),
        ('index.html', ".//li[@class='toctree-l3']/a", '1.2.1. Foo B1', True),
        ('index.html', ".//li[@class='toctree-l3']/a", '2.1.1. Bar A1', False),
        ('index.html', ".//li[@class='toctree-l3']/a", '2.2.1. Bar B1', False),
        # foo.html
        ('foo.html', './/h1', 'Foo', True),
        ('foo.html', './/h2', 'Foo A', True),
        ('foo.html', './/h3', 'Foo A1', True),
        ('foo.html', './/h2', 'Foo B', True),
        ('foo.html', './/h3', 'Foo B1', True),
        # foo.html section-number
        ('foo.html', ".//h1//span[@class='section-number']", '1. ', True),
        ('foo.html', ".//h2//span[@class='section-number']", '1.1. ', True),
        ('foo.html', ".//h3//span[@class='section-number']", '1.1.1. ', True),
        ('foo.html', ".//h2//span[@class='section-number']", '1.2. ', True),
        ('foo.html', ".//h3//span[@class='section-number']", '1.2.1. ', True),
        # foo.html sphinxsidebarwrapper
        ('foo.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '1.1. Foo A', True),
        (
            'foo.html',
            ".//div[@class='sphinxsidebarwrapper']//li/a",
            '1.1.1. Foo A1',
            True,
        ),
        ('foo.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '1.2. Foo B', True),
        (
            'foo.html',
            ".//div[@class='sphinxsidebarwrapper']//li/a",
            '1.2.1. Foo B1',
            True,
        ),
        # bar.html
        ('bar.html', './/h1', 'Bar', True),
        ('bar.html', './/h2', 'Bar A', True),
        ('bar.html', './/h2', 'Bar B', True),
        ('bar.html', './/h3', 'Bar B1', True),
        ('bar.html', ".//h1//span[@class='section-number']", '2. ', True),
        ('bar.html', ".//h2//span[@class='section-number']", '2.1. ', True),
        ('bar.html', ".//h2//span[@class='section-number']", '2.2. ', True),
        ('bar.html', ".//h3//span[@class='section-number']", '2.2.1. ', True),
        ('bar.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '2. Bar', True),
        ('bar.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '2.1. Bar A', True),
        ('bar.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '2.2. Bar B', True),
        (
            'bar.html',
            ".//div[@class='sphinxsidebarwrapper']//li/a",
            '2.2.1. Bar B1',
            False,
        ),
        # baz.html
        ('baz.html', './/h1', 'Baz A', True),
        ('baz.html', ".//h1//span[@class='section-number']", '2.1.1. ', True),
    ],
)
@pytest.mark.sphinx('html', testroot='tocdepth')
@pytest.mark.test_params(shared_result='test_build_html_tocdepth')
def test_tocdepth(app, cached_etree_parse, fname, path, check, be_found):
    app.build()
    # https://github.com/sphinx-doc/sphinx/issues/1251
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.parametrize(
    'expect',
    [
        (".//li[@class='toctree-l3']/a", '1.1.1. Foo A1', True),
        (".//li[@class='toctree-l3']/a", '1.2.1. Foo B1', True),
        (".//li[@class='toctree-l3']/a", '2.1.1. Bar A1', False),
        (".//li[@class='toctree-l3']/a", '2.2.1. Bar B1', False),
        (".//ul/li[@class='toctree-l1']/..//a", _intradocument_hyperlink_check),
        # index.rst
        ('.//h1', 'test-tocdepth', True),
        # foo.rst
        ('.//h2', 'Foo', True),
        ('.//h3', 'Foo A', True),
        ('.//h4', 'Foo A1', True),
        ('.//h3', 'Foo B', True),
        ('.//h4', 'Foo B1', True),
        (".//h2//span[@class='section-number']", '1. ', True),
        (".//h3//span[@class='section-number']", '1.1. ', True),
        (".//h4//span[@class='section-number']", '1.1.1. ', True),
        (".//h3//span[@class='section-number']", '1.2. ', True),
        (".//h4//span[@class='section-number']", '1.2.1. ', True),
        # bar.rst
        ('.//h2', 'Bar', True),
        ('.//h3', 'Bar A', True),
        ('.//h3', 'Bar B', True),
        ('.//h4', 'Bar B1', True),
        (".//h2//span[@class='section-number']", '2. ', True),
        (".//h3//span[@class='section-number']", '2.1. ', True),
        (".//h3//span[@class='section-number']", '2.2. ', True),
        (".//h4//span[@class='section-number']", '2.2.1. ', True),
        # baz.rst
        ('.//h4', 'Baz A', True),
        (".//h4//span[@class='section-number']", '2.1.1. ', True),
    ],
)
@pytest.mark.sphinx('singlehtml', testroot='tocdepth')
@pytest.mark.test_params(shared_result='test_build_html_tocdepth')
def test_tocdepth_singlehtml(app, cached_etree_parse, expect):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / 'index.html'), 'index.html', *expect)



from __future__ import annotations

from types import SimpleNamespace

from docutils import nodes
from docutils.parsers.rst.languages import en as english  # type: ignore[attr-defined]
from docutils.parsers.rst.states import (
    Inliner,
    RSTState,
    RSTStateMachine,
    state_classes,
)
from docutils.statemachine import StringList

from sphinx.util.docutils import SphinxDirective, new_document


def make_directive(
    *, env: SimpleNamespace, input_lines: StringList | None = None
) -> SphinxDirective:
    _, directive = make_directive_and_state(env=env, input_lines=input_lines)
    return directive


def make_directive_and_state(
    *, env: SimpleNamespace, input_lines: StringList | None = None
) -> tuple[RSTState, SphinxDirective]:
    sm = RSTStateMachine(state_classes, initial_state='Body')
    sm.reporter = object()
    if input_lines is not None:
        sm.input_lines = input_lines
    state = RSTState(sm)
    state.document = new_document('<tests>')
    state.document.settings.env = env
    state.document.settings.tab_width = 4
    state.document.settings.pep_references = None
    state.document.settings.rfc_references = None
    inliner = Inliner()
    inliner.init_customizations(state.document.settings)
    state.inliner = inliner
    state.parent = None
    state.memo = SimpleNamespace(
        document=state.document,
        language=english,
        inliner=state.inliner,
        reporter=state.document.reporter,
        section_level=0,
        title_styles=[],
    )
    directive = SphinxDirective(
        name='test_directive',
        arguments=[],
        options={},
        content=StringList(),
        lineno=0,
        content_offset=0,
        block_text='',
        state=state,
        state_machine=state.state_machine,
    )
    return state, directive


def test_sphinx_directive_env() -> None:
    state, directive = make_directive_and_state(env=SimpleNamespace())

    assert hasattr(directive, 'env')
    assert directive.env is state.document.settings.env


def test_sphinx_directive_config() -> None:
    env = SimpleNamespace(config=object())
    state, directive = make_directive_and_state(env=env)

    assert hasattr(directive, 'config')
    assert directive.config is directive.env.config
    assert directive.config is state.document.settings.env.config


def test_sphinx_directive_get_source_info() -> None:
    env = SimpleNamespace()
    input_lines = StringList(['spam'], source='<source>')
    directive = make_directive(env=env, input_lines=input_lines)

    assert directive.get_source_info() == ('<source>', 1)


def test_sphinx_directive_set_source_info() -> None:
    env = SimpleNamespace()
    input_lines = StringList(['spam'], source='<source>')
    directive = make_directive(env=env, input_lines=input_lines)

    node = nodes.Element()
    directive.set_source_info(node)
    assert node.source == '<source>'
    assert node.line == 1


def test_sphinx_directive_get_location() -> None:
    env = SimpleNamespace()
    input_lines = StringList(['spam'], source='<source>')
    directive = make_directive(env=env, input_lines=input_lines)

    assert directive.get_location() == '<source>:1'


def test_sphinx_directive_parse_content_to_nodes() -> None:
    directive = make_directive(env=SimpleNamespace())
    content = 'spam\n====\n\nEggs! *Lobster thermidor.*'
    directive.content = StringList(content.split('\n'), source='<source>')

    parsed = directive.parse_content_to_nodes(allow_section_headings=True)
    assert len(parsed) == 1
    node = parsed[0]
    assert isinstance(node, nodes.section)
    assert len(node.children) == 2
    assert isinstance(node.children[0], nodes.title)
    assert node.children[0].astext() == 'spam'
    assert isinstance(node.children[1], nodes.paragraph)
    assert node.children[1].astext() == 'Eggs! Lobster thermidor.'


def test_sphinx_directive_parse_text_to_nodes() -> None:
    directive = make_directive(env=SimpleNamespace())
    content = 'spam\n====\n\nEggs! *Lobster thermidor.*'

    parsed = directive.parse_text_to_nodes(content, allow_section_headings=True)
    assert len(parsed) == 1
    node = parsed[0]
    assert isinstance(node, nodes.section)
    assert len(node.children) == 2
    assert isinstance(node.children[0], nodes.title)
    assert node.children[0].astext() == 'spam'
    assert isinstance(node.children[1], nodes.paragraph)
    assert node.children[1].astext() == 'Eggs! Lobster thermidor.'


def test_sphinx_directive_parse_inline() -> None:
    directive = make_directive(env=SimpleNamespace())
    content = 'Eggs! *Lobster thermidor.*'

    parsed, messages = directive.parse_inline(content)
    assert len(parsed) == 2
    assert messages == []
    assert parsed[0] == nodes.Text('Eggs! ')
    assert isinstance(parsed[1], nodes.emphasis)
    assert parsed[1].rawsource == '*Lobster thermidor.*'
    assert parsed[1][0] == nodes.Text('Lobster thermidor.')



"""Tests the Python Domain"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sphinx import addnodes
from sphinx.addnodes import (
    desc,
    desc_addname,
    desc_annotation,
    desc_content,
    desc_name,
    desc_sig_keyword,
    desc_sig_space,
    desc_signature,
)
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx(
    'html',
    testroot='domain-py',
    freshenv=True,
)
def test_domain_py_canonical(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    content = (app.outdir / 'canonical.html').read_text(encoding='utf8')
    assert (
        '<a class="reference internal" href="#canonical.Foo" title="canonical.Foo">'
        '<code class="xref py py-class docutils literal notranslate">'
        '<span class="pre">Foo</span></code></a>'
    ) in content
    assert app.warning.getvalue() == ''


@pytest.mark.sphinx('html', testroot='_blank')
def test_canonical(app: SphinxTestApp) -> None:
    text = '.. py:class:: io.StringIO\n   :canonical: _io.StringIO'
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_addname, 'io.'],
                            [desc_name, 'StringIO'],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert 'io.StringIO' in domain.objects
    assert domain.objects['io.StringIO'] == ('index', 'io.StringIO', 'class', False)
    assert domain.objects['_io.StringIO'] == ('index', 'io.StringIO', 'class', True)


@pytest.mark.sphinx('html', testroot='_blank')
def test_canonical_definition_overrides(app: SphinxTestApp) -> None:
    text = (
        '.. py:class:: io.StringIO\n'
        '   :canonical: _io.StringIO\n'
        '.. py:class:: _io.StringIO\n'
    )
    restructuredtext.parse(app, text)
    assert app.warning.getvalue() == ''

    domain = app.env.domains.python_domain
    assert domain.objects['_io.StringIO'] == ('index', 'id0', 'class', False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_canonical_definition_skip(app: SphinxTestApp) -> None:
    text = (
        '.. py:class:: _io.StringIO\n'
        '.. py:class:: io.StringIO\n'
        '   :canonical: _io.StringIO\n'
    )

    restructuredtext.parse(app, text)
    assert app.warning.getvalue() == ''

    domain = app.env.domains.python_domain
    assert domain.objects['_io.StringIO'] == ('index', 'io.StringIO', 'class', False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_canonical_duplicated(app: SphinxTestApp) -> None:
    text = (
        '.. py:class:: mypackage.StringIO\n'
        '   :canonical: _io.StringIO\n'
        '.. py:class:: io.StringIO\n'
        '   :canonical: _io.StringIO\n'
    )

    restructuredtext.parse(app, text)
    assert app.warning.getvalue() != ''



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import ANY, call, patch

import pytest

from sphinx.builders.html import StandaloneHTMLBuilder

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='basic', confoverrides={'html_theme': 'alabaster'})
def test_html_pygments_style_default(app: SphinxTestApp) -> None:
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    style = app.builder.highlighter.formatter_args.get('style')
    assert style is not None
    assert style.__name__ == 'Alabaster'


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    confoverrides={'pygments_style': 'sphinx'},
)
def test_html_pygments_style_manually(app: SphinxTestApp) -> None:
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    style = app.builder.highlighter.formatter_args.get('style')
    assert style is not None
    assert style.__name__ == 'SphinxStyle'


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    confoverrides={'html_theme': 'classic'},
)
def test_html_pygments_for_classic_theme(app: SphinxTestApp) -> None:
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    style = app.builder.highlighter.formatter_args.get('style')
    assert style is not None
    assert style.__name__ == 'SphinxStyle'


@pytest.mark.sphinx('html', testroot='basic')
def test_html_dark_pygments_style_default(app: SphinxTestApp) -> None:
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    assert app.builder.dark_highlighter is None


@pytest.mark.sphinx('html', testroot='highlight_options')
def test_highlight_options(app: SphinxTestApp) -> None:
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    subject = app.builder.highlighter
    with patch.object(
        subject,
        'highlight_block',
        wraps=subject.highlight_block,
    ) as highlight:
        app.build()

        call_args = highlight.call_args_list
        assert len(call_args) == 3
        assert call_args[0] == call(
            ANY,
            'default',
            force=False,
            linenos=False,
            location=ANY,
            opts={'default_option': True},
        )
        assert call_args[1] == call(
            ANY,
            'python',
            force=False,
            linenos=False,
            location=ANY,
            opts={'python_option': True},
        )
        assert call_args[2] == call(
            ANY,
            'java',
            force=False,
            linenos=False,
            location=ANY,
            opts={},
        )


@pytest.mark.sphinx(
    'html',
    testroot='highlight_options',
    confoverrides={'highlight_options': {'default_option': True}},
)
def test_highlight_options_old(app: SphinxTestApp) -> None:
    assert isinstance(app.builder, StandaloneHTMLBuilder)  # type-checking
    subject = app.builder.highlighter
    with patch.object(
        subject,
        'highlight_block',
        wraps=subject.highlight_block,
    ) as highlight:
        app.build()

        call_args = highlight.call_args_list
        assert len(call_args) == 3
        assert call_args[0] == call(
            ANY,
            'default',
            force=False,
            linenos=False,
            location=ANY,
            opts={'default_option': True},
        )
        assert call_args[1] == call(
            ANY,
            'python',
            force=False,
            linenos=False,
            location=ANY,
            opts={},
        )
        assert call_args[2] == call(
            ANY,
            'java',
            force=False,
            linenos=False,
            location=ANY,
            opts={},
        )



"""Tests the Python Domain"""

from __future__ import annotations

import pytest
from docutils import nodes

from sphinx import addnodes
from sphinx.addnodes import (
    desc,
    desc_addname,
    desc_annotation,
    desc_content,
    desc_name,
    desc_parameterlist,
    desc_sig_keyword,
    desc_sig_punctuation,
    desc_sig_space,
    desc_signature,
    pending_xref,
)
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyexception_signature(app):
    text = '.. py:exception:: builtins.IOError'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'exception'], desc_sig_space),
                            ],
                            [desc_addname, 'builtins.'],
                            [desc_name, 'IOError'],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        desc,
        desctype='exception',
        domain='py',
        objtype='exception',
        no_index=False,
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_pydata_signature(app):
    text = '.. py:data:: version\n   :type: int\n   :value: 1\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'version'],
                            [
                                desc_annotation,
                                (
                                    [desc_sig_punctuation, ':'],
                                    desc_sig_space,
                                    [pending_xref, 'int'],
                                ),
                            ],
                            [
                                desc_annotation,
                                (
                                    desc_sig_space,
                                    [desc_sig_punctuation, '='],
                                    desc_sig_space,
                                    '1',
                                ),
                            ],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='data',
        domain='py',
        objtype='data',
        no_index=False,
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_pydata_signature_old(app):
    text = '.. py:data:: version\n   :annotation: = 1\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'version'],
                            [desc_annotation, (desc_sig_space, '= 1')],
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='data',
        domain='py',
        objtype='data',
        no_index=False,
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_pydata_with_union_type_operator(app):
    text = '.. py:data:: version\n   :type: int | str'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree[1][0],
        (
            [desc_name, 'version'],
            [
                desc_annotation,
                (
                    [desc_sig_punctuation, ':'],
                    desc_sig_space,
                    [pending_xref, 'int'],
                    desc_sig_space,
                    [desc_sig_punctuation, '|'],
                    desc_sig_space,
                    [pending_xref, 'str'],
                ),
            ],
        ),
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyobject_prefix(app):
    text = (
        '.. py:class:: Foo\n\n   .. py:method:: Foo.say\n   .. py:method:: FooBar.say'
    )
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_name, 'Foo'],
                        ),
                    ],
                    [desc_content, (addnodes.index, desc, addnodes.index, desc)],
                ),
            ],
        ),
    )
    # prefix is stripped
    assert doctree[1][1][1].astext().strip() == 'say()'
    # not stripped
    assert doctree[1][1][3].astext().strip() == 'FooBar.say()'


@pytest.mark.sphinx('html', testroot='_blank')
def test_pydata(app):
    text = '.. py:module:: example\n.. py:data:: var\n   :type: int\n'
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            addnodes.index,
            nodes.target,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_addname, 'example.'],
                            [desc_name, 'var'],
                            [
                                desc_annotation,
                                (
                                    [desc_sig_punctuation, ':'],
                                    desc_sig_space,
                                    [pending_xref, 'int'],
                                ),
                            ],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(doctree[3][0][2][2], pending_xref, **{'py:module': 'example'})
    assert 'example.var' in domain.objects
    assert domain.objects['example.var'] == ('index', 'example.var', 'data', False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyclass_options(app):
    text = '.. py:class:: Class1\n.. py:class:: Class2\n   :final:\n'
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_name, 'Class1'],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                (
                                    [desc_sig_keyword, 'final'],
                                    desc_sig_space,
                                    [desc_sig_keyword, 'class'],
                                    desc_sig_space,
                                ),
                            ],
                            [desc_name, 'Class2'],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )

    # class
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('single', 'Class1 (built-in class)', 'Class1', '', None)],
    )
    assert 'Class1' in domain.objects
    assert domain.objects['Class1'] == ('index', 'Class1', 'class', False)

    # :final:
    assert_node(
        doctree[2],
        addnodes.index,
        entries=[('single', 'Class2 (built-in class)', 'Class2', '', None)],
    )
    assert 'Class2' in domain.objects
    assert domain.objects['Class2'] == ('index', 'Class2', 'class', False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_pymethod_options(app):
    text = (
        '.. py:class:: Class\n'
        '\n'
        '   .. py:method:: meth1\n'
        '   .. py:method:: meth2\n'
        '      :classmethod:\n'
        '   .. py:method:: meth3\n'
        '      :staticmethod:\n'
        '   .. py:method:: meth4\n'
        '      :async:\n'
        '   .. py:method:: meth5\n'
        '      :abstractmethod:\n'
        '   .. py:method:: meth6\n'
        '      :final:\n'
    )
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_name, 'Class'],
                        ),
                    ],
                    [
                        desc_content,
                        (
                            addnodes.index,
                            desc,
                            addnodes.index,
                            desc,
                            addnodes.index,
                            desc,
                            addnodes.index,
                            desc,
                            addnodes.index,
                            desc,
                            addnodes.index,
                            desc,
                        ),
                    ],
                ),
            ],
        ),
    )

    # method
    assert_node(
        doctree[1][1][0],
        addnodes.index,
        entries=[('single', 'meth1() (Class method)', 'Class.meth1', '', None)],
    )
    assert_node(
        doctree[1][1][1],
        (
            [desc_signature, ([desc_name, 'meth1'], [desc_parameterlist, ()])],
            [desc_content, ()],
        ),
    )
    assert 'Class.meth1' in domain.objects
    assert domain.objects['Class.meth1'] == ('index', 'Class.meth1', 'method', False)

    # :classmethod:
    assert_node(
        doctree[1][1][2],
        addnodes.index,
        entries=[('single', 'meth2() (Class class method)', 'Class.meth2', '', None)],
    )
    assert_node(
        doctree[1][1][3],
        (
            [
                desc_signature,
                (
                    [
                        desc_annotation,
                        ([desc_sig_keyword, 'classmethod'], desc_sig_space),
                    ],
                    [desc_name, 'meth2'],
                    [desc_parameterlist, ()],
                ),
            ],
            [desc_content, ()],
        ),
    )
    assert 'Class.meth2' in domain.objects
    assert domain.objects['Class.meth2'] == ('index', 'Class.meth2', 'method', False)

    # :staticmethod:
    assert_node(
        doctree[1][1][4],
        addnodes.index,
        entries=[('single', 'meth3() (Class static method)', 'Class.meth3', '', None)],
    )
    assert_node(
        doctree[1][1][5],
        (
            [
                desc_signature,
                (
                    [desc_annotation, ([desc_sig_keyword, 'static'], desc_sig_space)],
                    [desc_name, 'meth3'],
                    [desc_parameterlist, ()],
                ),
            ],
            [desc_content, ()],
        ),
    )
    assert 'Class.meth3' in domain.objects
    assert domain.objects['Class.meth3'] == ('index', 'Class.meth3', 'method', False)

    # :async:
    assert_node(
        doctree[1][1][6],
        addnodes.index,
        entries=[('single', 'meth4() (Class method)', 'Class.meth4', '', None)],
    )
    assert_node(
        doctree[1][1][7],
        (
            [
                desc_signature,
                (
                    [desc_annotation, ([desc_sig_keyword, 'async'], desc_sig_space)],
                    [desc_name, 'meth4'],
                    [desc_parameterlist, ()],
                ),
            ],
            [desc_content, ()],
        ),
    )
    assert 'Class.meth4' in domain.objects
    assert domain.objects['Class.meth4'] == ('index', 'Class.meth4', 'method', False)

    # :abstractmethod:
    assert_node(
        doctree[1][1][8],
        addnodes.index,
        entries=[('single', 'meth5() (Class method)', 'Class.meth5', '', None)],
    )
    assert_node(
        doctree[1][1][9],
        (
            [
                desc_signature,
                (
                    [
                        desc_annotation,
                        ([desc_sig_keyword, 'abstractmethod'], desc_sig_space),
                    ],
                    [desc_name, 'meth5'],
                    [desc_parameterlist, ()],
                ),
            ],
            [desc_content, ()],
        ),
    )
    assert 'Class.meth5' in domain.objects
    assert domain.objects['Class.meth5'] == ('index', 'Class.meth5', 'method', False)

    # :final:
    assert_node(
        doctree[1][1][10],
        addnodes.index,
        entries=[('single', 'meth6() (Class method)', 'Class.meth6', '', None)],
    )
    assert_node(
        doctree[1][1][11],
        (
            [
                desc_signature,
                (
                    [desc_annotation, ([desc_sig_keyword, 'final'], desc_sig_space)],
                    [desc_name, 'meth6'],
                    [desc_parameterlist, ()],
                ),
            ],
            [desc_content, ()],
        ),
    )
    assert 'Class.meth6' in domain.objects
    assert domain.objects['Class.meth6'] == ('index', 'Class.meth6', 'method', False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyclassmethod(app):
    text = '.. py:class:: Class\n\n   .. py:classmethod:: meth\n'
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_name, 'Class'],
                        ),
                    ],
                    [desc_content, (addnodes.index, desc)],
                ),
            ],
        ),
    )
    assert_node(
        doctree[1][1][0],
        addnodes.index,
        entries=[('single', 'meth() (Class class method)', 'Class.meth', '', None)],
    )
    assert_node(
        doctree[1][1][1],
        (
            [
                desc_signature,
                (
                    [
                        desc_annotation,
                        ([desc_sig_keyword, 'classmethod'], desc_sig_space),
                    ],
                    [desc_name, 'meth'],
                    [desc_parameterlist, ()],
                ),
            ],
            [desc_content, ()],
        ),
    )
    assert 'Class.meth' in domain.objects
    assert domain.objects['Class.meth'] == ('index', 'Class.meth', 'method', False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_pystaticmethod(app):
    text = '.. py:class:: Class\n\n   .. py:staticmethod:: meth\n'
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_name, 'Class'],
                        ),
                    ],
                    [desc_content, (addnodes.index, desc)],
                ),
            ],
        ),
    )
    assert_node(
        doctree[1][1][0],
        addnodes.index,
        entries=[('single', 'meth() (Class static method)', 'Class.meth', '', None)],
    )
    assert_node(
        doctree[1][1][1],
        (
            [
                desc_signature,
                (
                    [desc_annotation, ([desc_sig_keyword, 'static'], desc_sig_space)],
                    [desc_name, 'meth'],
                    [desc_parameterlist, ()],
                ),
            ],
            [desc_content, ()],
        ),
    )
    assert 'Class.meth' in domain.objects
    assert domain.objects['Class.meth'] == ('index', 'Class.meth', 'method', False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyattribute(app):
    text = (
        '.. py:class:: Class\n'
        '\n'
        '   .. py:attribute:: attr\n'
        '      :type: Optional[str]\n'
        "      :value: ''\n"
    )
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_name, 'Class'],
                        ),
                    ],
                    [desc_content, (addnodes.index, desc)],
                ),
            ],
        ),
    )
    assert_node(
        doctree[1][1][0],
        addnodes.index,
        entries=[('single', 'attr (Class attribute)', 'Class.attr', '', None)],
    )
    assert_node(
        doctree[1][1][1],
        (
            [
                desc_signature,
                (
                    [desc_name, 'attr'],
                    [
                        desc_annotation,
                        (
                            [desc_sig_punctuation, ':'],
                            desc_sig_space,
                            [pending_xref, 'str'],
                            desc_sig_space,
                            [desc_sig_punctuation, '|'],
                            desc_sig_space,
                            [pending_xref, 'None'],
                        ),
                    ],
                    [
                        desc_annotation,
                        (
                            desc_sig_space,
                            [desc_sig_punctuation, '='],
                            desc_sig_space,
                            "''",
                        ),
                    ],
                ),
            ],
            [desc_content, ()],
        ),
    )
    assert_node(doctree[1][1][1][0][1][2], pending_xref, **{'py:class': 'Class'})
    assert_node(doctree[1][1][1][0][1][6], pending_xref, **{'py:class': 'Class'})
    assert 'Class.attr' in domain.objects
    assert domain.objects['Class.attr'] == ('index', 'Class.attr', 'attribute', False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_pyproperty(app):
    text = (
        '.. py:class:: Class\n'
        '\n'
        '   .. py:property:: prop1\n'
        '      :abstractmethod:\n'
        '      :type: str\n'
        '\n'
        '   .. py:property:: prop2\n'
        '      :classmethod:\n'
        '      :type: str\n'
    )
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_name, 'Class'],
                        ),
                    ],
                    [desc_content, (addnodes.index, desc, addnodes.index, desc)],
                ),
            ],
        ),
    )
    assert_node(
        doctree[1][1][0],
        addnodes.index,
        entries=[('single', 'prop1 (Class property)', 'Class.prop1', '', None)],
    )
    assert_node(
        doctree[1][1][1],
        (
            [
                desc_signature,
                (
                    [
                        desc_annotation,
                        (
                            [desc_sig_keyword, 'abstract'],
                            desc_sig_space,
                            [desc_sig_keyword, 'property'],
                            desc_sig_space,
                        ),
                    ],
                    [desc_name, 'prop1'],
                    [
                        desc_annotation,
                        (
                            [desc_sig_punctuation, ':'],
                            desc_sig_space,
                            [pending_xref, 'str'],
                        ),
                    ],
                ),
            ],
            [desc_content, ()],
        ),
    )
    assert_node(
        doctree[1][1][2],
        addnodes.index,
        entries=[('single', 'prop2 (Class property)', 'Class.prop2', '', None)],
    )
    assert_node(
        doctree[1][1][3],
        (
            [
                desc_signature,
                (
                    [
                        desc_annotation,
                        (
                            [desc_sig_keyword, 'class'],
                            desc_sig_space,
                            [desc_sig_keyword, 'property'],
                            desc_sig_space,
                        ),
                    ],
                    [desc_name, 'prop2'],
                    [
                        desc_annotation,
                        (
                            [desc_sig_punctuation, ':'],
                            desc_sig_space,
                            [pending_xref, 'str'],
                        ),
                    ],
                ),
            ],
            [desc_content, ()],
        ),
    )
    assert 'Class.prop1' in domain.objects
    assert domain.objects['Class.prop1'] == ('index', 'Class.prop1', 'property', False)
    assert 'Class.prop2' in domain.objects
    assert domain.objects['Class.prop2'] == ('index', 'Class.prop2', 'property', False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_py_type_alias(app):
    text = (
        '.. py:module:: example\n'
        '.. py:type:: Alias1\n'
        '   :canonical: list[str | int]\n'
        '\n'
        '.. py:class:: Class\n'
        '\n'
        '   .. py:type:: Alias2\n'
        '      :canonical: int\n'
    )
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            addnodes.index,
            nodes.target,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'type'], desc_sig_space),
                            ],
                            [desc_addname, 'example.'],
                            [desc_name, 'Alias1'],
                            [
                                desc_annotation,
                                (
                                    desc_sig_space,
                                    [desc_sig_punctuation, '='],
                                    desc_sig_space,
                                    [pending_xref, 'list'],
                                    [desc_sig_punctuation, '['],
                                    [pending_xref, 'str'],
                                    desc_sig_space,
                                    [desc_sig_punctuation, '|'],
                                    desc_sig_space,
                                    [pending_xref, 'int'],
                                    [desc_sig_punctuation, ']'],
                                ),
                            ],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_addname, 'example.'],
                            [desc_name, 'Class'],
                        ),
                    ],
                    [desc_content, (addnodes.index, desc)],
                ),
            ],
        ),
    )
    assert_node(
        doctree[5][1][0],
        addnodes.index,
        entries=[
            (
                'single',
                'Alias2 (type alias in example.Class)',
                'example.Class.Alias2',
                '',
                None,
            )
        ],
    )
    assert_node(
        doctree[5][1][1],
        (
            [
                desc_signature,
                (
                    [desc_annotation, ([desc_sig_keyword, 'type'], desc_sig_space)],
                    [desc_name, 'Alias2'],
                    [
                        desc_annotation,
                        (
                            desc_sig_space,
                            [desc_sig_punctuation, '='],
                            desc_sig_space,
                            [pending_xref, 'int'],
                        ),
                    ],
                ),
            ],
            [desc_content, ()],
        ),
    )
    assert 'example.Alias1' in domain.objects
    assert domain.objects['example.Alias1'] == (
        'index',
        'example.Alias1',
        'type',
        False,
    )
    assert 'example.Class.Alias2' in domain.objects
    assert domain.objects['example.Class.Alias2'] == (
        'index',
        'example.Class.Alias2',
        'type',
        False,
    )


@pytest.mark.sphinx('html', testroot='domain-py', freshenv=True)
def test_domain_py_type_alias(app):
    app.build(force_all=True)

    content = (app.outdir / 'type_alias.html').read_text(encoding='utf8')
    assert (
        '<em class="property"><span class="k"><span class="pre">type</span></span><span class="w"> </span></em>'
        '<span class="sig-prename descclassname"><span class="pre">module_one.</span></span>'
        '<span class="sig-name descname"><span class="pre">MyAlias</span></span>'
        '<em class="property"><span class="w"> </span><span class="p"><span class="pre">=</span></span>'
        '<span class="w"> </span><span class="pre">list</span>'
        '<span class="p"><span class="pre">[</span></span>'
        '<span class="pre">int</span><span class="w"> </span>'
        '<span class="p"><span class="pre">|</span></span><span class="w"> </span>'
        '<a class="reference internal" href="#module_two.SomeClass" title="module_two.SomeClass">'
        '<span class="pre">module_two.SomeClass</span></a>'
        '<span class="p"><span class="pre">]</span></span></em>'
    ) in content
    assert app.warning.getvalue() == ''


@pytest.mark.sphinx('html', testroot='_blank')
def test_pydecorator_signature(app):
    text = '.. py:decorator:: deco'
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [desc_signature, ([desc_addname, '@'], [desc_name, 'deco'])],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='function',
        domain='py',
        objtype='function',
        no_index=False,
    )

    assert 'deco' in domain.objects
    assert domain.objects['deco'] == ('index', 'deco', 'function', False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_pydecoratormethod_signature(app):
    text = '.. py:decoratormethod:: deco'
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [desc_signature, ([desc_addname, '@'], [desc_name, 'deco'])],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='method',
        domain='py',
        objtype='method',
        no_index=False,
    )

    assert 'deco' in domain.objects
    assert domain.objects['deco'] == ('index', 'deco', 'method', False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_pycurrentmodule(app):
    text = (
        '.. py:module:: Other\n'
        '\n'
        '.. py:module:: Module\n'
        '.. py:class:: A\n'
        '\n'
        '   .. py:method:: m1\n'
        '   .. py:method:: m2\n'
        '\n'
        '.. py:currentmodule:: Other\n'
        '\n'
        '.. py:class:: B\n'
        '\n'
        '   .. py:method:: m3\n'
        '   .. py:method:: m4\n'
    )
    domain = app.env.domains.python_domain
    doctree = restructuredtext.parse(app, text)
    print(doctree)
    assert_node(
        doctree,
        (
            addnodes.index,
            addnodes.index,
            addnodes.index,
            nodes.target,
            nodes.target,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_addname, 'Module.'],
                            [desc_name, 'A'],
                        ),
                    ],
                    [
                        desc_content,
                        (
                            addnodes.index,
                            [
                                desc,
                                (
                                    [
                                        desc_signature,
                                        (
                                            [desc_name, 'm1'],
                                            [desc_parameterlist, ()],
                                        ),
                                    ],
                                    [desc_content, ()],
                                ),
                            ],
                            addnodes.index,
                            [
                                desc,
                                (
                                    [
                                        desc_signature,
                                        (
                                            [desc_name, 'm2'],
                                            [desc_parameterlist, ()],
                                        ),
                                    ],
                                    [desc_content, ()],
                                ),
                            ],
                        ),
                    ],
                ),
            ],
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_addname, 'Other.'],
                            [desc_name, 'B'],
                        ),
                    ],
                    [
                        desc_content,
                        (
                            addnodes.index,
                            [
                                desc,
                                (
                                    [
                                        desc_signature,
                                        (
                                            [desc_name, 'm3'],
                                            [desc_parameterlist, ()],
                                        ),
                                    ],
                                    [desc_content, ()],
                                ),
                            ],
                            addnodes.index,
                            [
                                desc,
                                (
                                    [
                                        desc_signature,
                                        (
                                            [desc_name, 'm4'],
                                            [desc_parameterlist, ()],
                                        ),
                                    ],
                                    [desc_content, ()],
                                ),
                            ],
                        ),
                    ],
                ),
            ],
        ),
    )
    assert 'Module' in domain.objects
    assert domain.objects['Module'] == ('index', 'module-Module', 'module', False)
    assert 'Other' in domain.objects
    assert domain.objects['Other'] == ('index', 'module-Other', 'module', False)
    assert 'Module.A' in domain.objects
    assert domain.objects['Module.A'] == ('index', 'Module.A', 'class', False)
    assert 'Other.B' in domain.objects
    assert domain.objects['Other.B'] == ('index', 'Other.B', 'class', False)
    assert 'Module.A.m1' in domain.objects
    assert domain.objects['Module.A.m1'] == ('index', 'Module.A.m1', 'method', False)
    assert 'Module.A.m2' in domain.objects
    assert domain.objects['Module.A.m2'] == ('index', 'Module.A.m2', 'method', False)
    assert 'Other.B.m3' in domain.objects
    assert domain.objects['Other.B.m3'] == ('index', 'Other.B.m3', 'method', False)
    assert 'Other.B.m4' in domain.objects
    assert domain.objects['Other.B.m4'] == ('index', 'Other.B.m4', 'method', False)



"""Test the search index builder."""

from __future__ import annotations

import json
import warnings
from io import BytesIO
from typing import TYPE_CHECKING

import pytest
from docutils import frontend, utils
from docutils.parsers import rst

from sphinx.search import IndexBuilder

from tests.utils import TESTS_ROOT

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from pathlib import Path
    from typing import Any

    from sphinx.domains import ObjType
    from sphinx.testing.util import SphinxTestApp

JAVASCRIPT_TEST_ROOTS = [
    directory
    for directory in (TESTS_ROOT / 'js' / 'roots').iterdir()
    if (directory / 'conf.py').exists()
]


class DummyDomainsContainer:
    def __init__(self, **domains: DummyDomain) -> None:
        self._domain_instances = domains

    def sorted(self) -> Iterator[DummyDomain]:
        for _domain_name, domain in sorted(self._domain_instances.items()):
            yield domain


class DummyEnvironment:
    def __init__(self, version: str, domains: DummyDomainsContainer) -> None:
        self.version = version
        self.domains = domains

    def __getattr__(self, name: str):
        if name.startswith('_search_index_'):
            setattr(self, name, {})
        return getattr(self, name, {})

    def __str__(self) -> str:
        return f'DummyEnvironment({self.version!r}, {self.domains!r})'


class DummyDomain:
    def __init__(
        self, name: str, data: Iterable[tuple[str, str, str, str, str, int]]
    ) -> None:
        self.name = name
        self.data = data
        self.object_types: dict[str, ObjType] = {}

    def get_objects(self) -> Iterable[tuple[str, str, str, str, str, int]]:
        return self.data


def load_searchindex(path: Path) -> Any:
    searchindex = path.read_text(encoding='utf8')
    assert searchindex.startswith('Search.setIndex(')
    assert searchindex.endswith(')')

    return json.loads(searchindex[16:-1])


def is_registered_term(index: Any, keyword: str) -> bool:
    return index['terms'].get(keyword, []) != []


FILE_CONTENTS = """\
section_title
=============

.. test that comments are not indexed: boson

another_title
=============

test that non-comments are indexed: fermion
"""


@pytest.mark.sphinx('html', testroot='ext-viewcode')
def test_objects_are_escaped(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    index = load_searchindex(app.outdir / 'searchindex.js')
    for item in index.get('objects').get(''):
        if item[-1] == 'n::Array&lt;T, d&gt;':  # n::Array<T,d> is escaped
            break
    else:
        raise AssertionError(index.get('objects').get(''))


@pytest.mark.sphinx('html', testroot='search')
def test_meta_keys_are_handled_for_language_en(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    searchindex = load_searchindex(app.outdir / 'searchindex.js')
    assert not is_registered_term(searchindex, 'thisnoteith')
    assert is_registered_term(searchindex, 'thisonetoo')
    assert is_registered_term(searchindex, 'findthiskei')
    assert is_registered_term(searchindex, 'thistoo')
    assert not is_registered_term(searchindex, 'onlygerman')
    assert is_registered_term(searchindex, 'notgerman')
    assert not is_registered_term(searchindex, 'onlytoogerman')


@pytest.mark.sphinx(
    'html',
    testroot='search',
    confoverrides={'html_search_language': 'de'},
    freshenv=True,
)
def test_meta_keys_are_handled_for_language_de(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    searchindex = load_searchindex(app.outdir / 'searchindex.js')
    assert not is_registered_term(searchindex, 'thisnoteith')
    assert is_registered_term(searchindex, 'thisonetoo')
    assert not is_registered_term(searchindex, 'findthiskei')
    assert not is_registered_term(searchindex, 'thistoo')
    assert is_registered_term(searchindex, 'onlygerman')
    assert not is_registered_term(searchindex, 'notgerman')
    assert is_registered_term(searchindex, 'onlytoogerman')


@pytest.mark.sphinx('html', testroot='search')
def test_stemmer_does_not_remove_short_words(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    searchindex = (app.outdir / 'searchindex.js').read_text(encoding='utf8')
    assert 'bat' in searchindex


@pytest.mark.sphinx('html', testroot='search')
def test_stemmer(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    searchindex = load_searchindex(app.outdir / 'searchindex.js')
    print(searchindex)
    assert is_registered_term(searchindex, 'findthisstemmedkei')
    assert is_registered_term(searchindex, 'intern')


@pytest.mark.sphinx('html', testroot='search')
def test_term_in_heading_and_section(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    searchindex = (app.outdir / 'searchindex.js').read_text(encoding='utf8')
    # if search term is in the title of one doc and in the text of another
    # both documents should be a hit in the search index as a title,
    # respectively text hit
    assert '"textinhead":2' in searchindex
    assert '"textinhead":0' in searchindex


@pytest.mark.sphinx('html', testroot='search')
def test_term_in_raw_directive(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    searchindex = load_searchindex(app.outdir / 'searchindex.js')
    assert not is_registered_term(searchindex, 'raw')
    assert is_registered_term(searchindex, 'rawword')
    assert not is_registered_term(searchindex, 'latex_keyword')


def test_IndexBuilder():
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        # DeprecationWarning: The frontend.OptionParser class will be replaced
        # by a subclass of argparse.ArgumentParser in Docutils 0.21 or later.
        optparser = frontend.OptionParser(components=(rst.Parser,))
    settings = optparser.get_default_values()
    parser = rst.Parser()

    domain1 = DummyDomain(
        'dummy1',
        [
            ('objname1', 'objdispname1', 'objtype1', 'docname1_1', '#anchor', 1),
            ('objname2', 'objdispname2', 'objtype2', 'docname1_2', '', -1),
        ],
    )
    domain2 = DummyDomain(
        'dummy2',
        [
            ('objname1', 'objdispname1', 'objtype1', 'docname2_1', '#anchor', 1),
            ('objname2', 'objdispname2', 'objtype2', 'docname2_2', '', -1),
        ],
    )
    env = DummyEnvironment('1.0', DummyDomainsContainer(dummy1=domain1, dummy2=domain2))
    doc = utils.new_document('test data', settings)
    doc['file'] = 'dummy'
    parser.parse(FILE_CONTENTS, doc)

    # feed
    index = IndexBuilder(env, 'en', {}, '')
    index.feed('docname1_1', 'filename1_1', 'title1_1', doc)
    index.feed('docname1_2', 'filename1_2', 'title1_2', doc)
    index.feed('docname2_2', 'filename2_2', 'title2_2', doc)
    index.feed('docname2_1', 'filename2_1', 'title2_1', doc)
    assert index._titles == {
        'docname1_1': 'title1_1',
        'docname1_2': 'title1_2',
        'docname2_1': 'title2_1',
        'docname2_2': 'title2_2',
    }
    assert index._filenames == {
        'docname1_1': 'filename1_1',
        'docname1_2': 'filename1_2',
        'docname2_1': 'filename2_1',
        'docname2_2': 'filename2_2',
    }
    # note: element iteration order (sort order) is important when the index
    # is frozen (serialized) during build -- however, the _mapping-related
    # dictionaries below may be iterated in arbitrary order by Python at
    # runtime.
    assert index._mapping == {
        'ar': {'docname1_1', 'docname1_2', 'docname2_1', 'docname2_2'},
        'fermion': {'docname1_1', 'docname1_2', 'docname2_1', 'docname2_2'},
        'comment': {'docname1_1', 'docname1_2', 'docname2_1', 'docname2_2'},
        'non': {'docname1_1', 'docname1_2', 'docname2_1', 'docname2_2'},
        'index': {'docname1_1', 'docname1_2', 'docname2_1', 'docname2_2'},
        'test': {'docname1_1', 'docname1_2', 'docname2_1', 'docname2_2'},
    }
    assert index._title_mapping == {
        'another_titl': {'docname1_1', 'docname1_2', 'docname2_1', 'docname2_2'},
        'section_titl': {'docname1_1', 'docname1_2', 'docname2_1', 'docname2_2'},
    }
    assert index._objtypes == {}
    assert index._objnames == {}

    # freeze
    assert index.freeze() == {
        'docnames': ('docname1_1', 'docname1_2', 'docname2_1', 'docname2_2'),
        'envversion': '1.0',
        'filenames': ['filename1_1', 'filename1_2', 'filename2_1', 'filename2_2'],
        'objects': {
            '': [
                (0, 0, 1, '#anchor', 'objdispname1'),
                (2, 1, 1, '#anchor', 'objdispname1'),
            ]
        },
        'objnames': {
            0: ('dummy1', 'objtype1', 'objtype1'),
            1: ('dummy2', 'objtype1', 'objtype1'),
        },
        'objtypes': {0: 'dummy1:objtype1', 1: 'dummy2:objtype1'},
        'terms': {
            'ar': [0, 1, 2, 3],
            'comment': [0, 1, 2, 3],
            'fermion': [0, 1, 2, 3],
            'index': [0, 1, 2, 3],
            'non': [0, 1, 2, 3],
            'test': [0, 1, 2, 3],
        },
        'titles': ('title1_1', 'title1_2', 'title2_1', 'title2_2'),
        'titleterms': {
            'another_titl': [0, 1, 2, 3],
            'section_titl': [0, 1, 2, 3],
        },
        'alltitles': {
            'another_title': [
                (0, 'another-title'),
                (1, 'another-title'),
                (2, 'another-title'),
                (3, 'another-title'),
            ],
            'section_title': [(0, None), (1, None), (2, None), (3, None)],
        },
        'indexentries': {},
    }
    assert index._objtypes == {('dummy1', 'objtype1'): 0, ('dummy2', 'objtype1'): 1}
    assert index._objnames == {
        0: ('dummy1', 'objtype1', 'objtype1'),
        1: ('dummy2', 'objtype1', 'objtype1'),
    }

    env = DummyEnvironment('1.0', DummyDomainsContainer(dummy1=domain1, dummy2=domain2))

    # dump / load
    stream = BytesIO()
    index.dump(stream, 'pickle')
    stream.seek(0)

    index2 = IndexBuilder(env, 'en', {}, '')
    index2.load(stream, 'pickle')

    assert index2._titles == index._titles
    assert index2._filenames == index._filenames
    assert index2._mapping == index._mapping
    assert index2._title_mapping == index._title_mapping
    assert index2._objtypes == {}
    assert index2._objnames == {}

    # freeze after load
    assert index2.freeze() == index.freeze()
    assert index2._objtypes == index._objtypes
    assert index2._objnames == index._objnames

    # prune
    index.prune(['docname1_2', 'docname2_2'])
    assert index._titles == {'docname1_2': 'title1_2', 'docname2_2': 'title2_2'}
    assert index._filenames == {
        'docname1_2': 'filename1_2',
        'docname2_2': 'filename2_2',
    }
    assert index._mapping == {
        'ar': {'docname1_2', 'docname2_2'},
        'fermion': {'docname1_2', 'docname2_2'},
        'comment': {'docname1_2', 'docname2_2'},
        'non': {'docname1_2', 'docname2_2'},
        'index': {'docname1_2', 'docname2_2'},
        'test': {'docname1_2', 'docname2_2'},
    }
    assert index._title_mapping == {
        'another_titl': {'docname1_2', 'docname2_2'},
        'section_titl': {'docname1_2', 'docname2_2'},
    }
    assert index._objtypes == {('dummy1', 'objtype1'): 0, ('dummy2', 'objtype1'): 1}
    assert index._objnames == {
        0: ('dummy1', 'objtype1', 'objtype1'),
        1: ('dummy2', 'objtype1', 'objtype1'),
    }

    # freeze after prune
    assert index.freeze() == {
        'docnames': ('docname1_2', 'docname2_2'),
        'envversion': '1.0',
        'filenames': ['filename1_2', 'filename2_2'],
        'objects': {},
        'objnames': {
            0: ('dummy1', 'objtype1', 'objtype1'),
            1: ('dummy2', 'objtype1', 'objtype1'),
        },
        'objtypes': {0: 'dummy1:objtype1', 1: 'dummy2:objtype1'},
        'terms': {
            'ar': [0, 1],
            'comment': [0, 1],
            'fermion': [0, 1],
            'index': [0, 1],
            'non': [0, 1],
            'test': [0, 1],
        },
        'titles': ('title1_2', 'title2_2'),
        'titleterms': {
            'another_titl': [0, 1],
            'section_titl': [0, 1],
        },
        'alltitles': {
            'another_title': [(0, 'another-title'), (1, 'another-title')],
            'section_title': [(0, None), (1, None)],
        },
        'indexentries': {},
    }
    assert index._objtypes == {('dummy1', 'objtype1'): 0, ('dummy2', 'objtype1'): 1}
    assert index._objnames == {
        0: ('dummy1', 'objtype1', 'objtype1'),
        1: ('dummy2', 'objtype1', 'objtype1'),
    }


def test_IndexBuilder_lookup():
    env = DummyEnvironment('1.0', {})

    # zh
    index = IndexBuilder(env, 'zh', {}, '')
    assert index.lang.lang == 'zh'

    # zh_CN
    index = IndexBuilder(env, 'zh_CN', {}, '')
    assert index.lang.lang == 'zh'


@pytest.mark.sphinx(
    'html',
    testroot='search',
    confoverrides={'html_search_language': 'zh'},
    srcdir='search_zh',
)
def test_search_index_gen_zh(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    index = load_searchindex(app.outdir / 'searchindex.js')
    assert 'chinesetest ' not in index['terms']
    assert 'chinesetest' in index['terms']
    assert 'chinesetesttwo' in index['terms']
    assert 'cas' in index['terms']


@pytest.mark.sphinx(
    'html',
    testroot='search',
    freshenv=True,
)
def test_nosearch(app: SphinxTestApp) -> None:
    app.build()
    index = load_searchindex(app.outdir / 'searchindex.js')
    assert index['docnames'] == ['index', 'nosearch', 'tocitem']
    # latex is in 'nosearch.rst', and nowhere else
    assert 'latex' not in index['terms']
    # cat is in 'index.rst' but is marked with the 'no-search' class
    assert 'cat' not in index['terms']
    # bat is indexed from 'index.rst' and 'tocitem.rst' (document IDs 0, 2), and
    # not from 'nosearch.rst' (document ID 1)
    assert 'bat' in index['terms']
    assert index['terms']['bat'] == [0, 2]


@pytest.mark.sphinx(
    'html',
    testroot='search',
    parallel=3,
    freshenv=True,
)
def test_parallel(app: SphinxTestApp) -> None:
    app.build()
    index = load_searchindex(app.outdir / 'searchindex.js')
    assert index['docnames'] == ['index', 'nosearch', 'tocitem']


@pytest.mark.sphinx('html', testroot='search')
def test_search_index_is_deterministic(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    index = load_searchindex(app.outdir / 'searchindex.js')
    # Pretty print the index. Only shown by pytest on failure.
    print(f'searchindex.js contents:\n\n{json.dumps(index, indent=2)}')
    assert_is_sorted(index, '')


def is_title_tuple_type(item: list[int | str]) -> bool:
    """In the search index, titles inside .alltitles are stored as a tuple of
    (document_idx, title_anchor). Tuples are represented as lists in JSON,
    but their contents must not be sorted. We cannot sort them anyway, as
    document_idx is an int and title_anchor is a str.
    """
    return len(item) == 2 and isinstance(item[0], int) and isinstance(item[1], str)


def assert_is_sorted(
    item: dict[str, str] | list[int | str] | int | str, path: str
) -> None:
    lists_not_to_sort = {
        # Each element of .titles is related to the element of .docnames in the same position.
        # The ordering is deterministic because .docnames is sorted.
        '.titles',
        # Each element of .filenames is related to the element of .docnames in the same position.
        # The ordering is deterministic because .docnames is sorted.
        '.filenames',
    }

    err_path = path or '<root>'
    if isinstance(item, dict):
        assert list(item.keys()) == sorted(item.keys()), f'{err_path} is not sorted'
        for key, value in item.items():
            assert_is_sorted(value, f'{path}.{key}')
    elif isinstance(item, list):
        if not is_title_tuple_type(item) and path not in lists_not_to_sort:
            # sort nulls last; http://stackoverflow.com/questions/19868767/
            assert item == sorted(item, key=lambda x: (x is None, x)), (
                f'{err_path} is not sorted'
            )
        for i, child in enumerate(item):
            assert_is_sorted(child, f'{path}[{i}]')


@pytest.mark.parametrize('directory', JAVASCRIPT_TEST_ROOTS)
def test_check_js_search_indexes(make_app, sphinx_test_tempdir, directory):
    app = make_app(
        'html',
        srcdir=directory,
        builddir=sphinx_test_tempdir / directory.name,
    )
    app.build()

    fresh_searchindex = app.outdir / 'searchindex.js'
    existing_searchindex = (
        TESTS_ROOT / 'js' / 'fixtures' / directory.name / 'searchindex.js'
    )

    msg = (
        f'Search index fixture {existing_searchindex} does not match regenerated copy.'
    )
    assert fresh_searchindex.read_bytes() == existing_searchindex.read_bytes(), msg



import sys
from pathlib import Path

sys.path.insert(0, str(Path.cwd().resolve()))

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.autosummary']
autodoc_default_options = {'members': True}
autosummary_ignore_module_all = False
autosummary_imorted_members = False

templates_path = ['templates']



class Foo:
    pass


class DocHere(Foo):
    pass


class DocSubDir1(DocHere):
    pass


class DocSubDir2(DocSubDir1):
    pass


class DocMainLevel(Foo):
    pass


class Alice(object):  # NoQA: UP004
    pass



from __future__ import annotations

from types import SimpleNamespace

from docutils import nodes

from sphinx.transforms.post_transforms.images import ImageConverter
from sphinx.util.docutils import new_document

WEBP_DATA = (
    b'RIFF\xa8\x01\x00\x00WEBPVP8X\n\x00\x00\x00'
    b"\x10\x00\x00\x00\x0f\x00\x00\x0f\x00\x00ALPH\xc3\x00\x00\x00\x01'"
    b'\xa2\xa8\x91$\xe5z\xe7\x18_\xe7\xdf*\x99\x88\x98\xfftq\x8d\xe0'
    b'&0\xe2\xe1\x8bw2\xc8\xc1\x11\\\x83+0\xe8\xb0x\x15\x8ex'
    b'Q5\xc1\x08\x0c\x02O\x92\xa0j\xb0U\x19\x1c\xd6\xb6mF/N'
    b'\xc6v<\xb6\xedw\xfb\xaf)\xae!\xa2\xffI\xd1\xfd\x8f\x90\xf7\xba'
    b'DI$\x1b:%\x914\xf3\x14m\x0e\xc7\xd3\xe5\x16 \xf4\x0b\x14'
    b'\xbe\x90\xe1\x83\xb7\x1a2\x9e6\x82\x7f\x1d)~Nv\x08\xfb\x88\x9e'
    b'\xb3\x91\xef\x99sF\xe82\x82\xdb\xf8\xccH\xb2\xf7E0} \xfd'
    b'6\x17\x8c!2V-\xa5\xd6k#\xbc]\xe3\xa5Y\x15\xd5\x9c\x81'
    b'\xa4\xd9n\x96u\x8a\x181\x0f\x8a\xaa,P4\xfa0\x82\xdf\xbak'
    b'PR)\xb5-\xcf\xe9T\x14\n\x01\x00\x00\x00VP8 \xbe\x00'
    b'\x00\x00\x90\x02\x00\x9d\x01*\x10\x00\x10\x00\x03\x004%\xb0\x02t0'
    b'O\x08\x85\x0c|\x03\x1d\x08,\xfd\xe8\x00\xfe\xfdt\xa0\xfd\x02\x9b\x1f'
    b'\x8a\xf7C|\x9c7\xf6\xd2\x0c\xaf\xd3\xff5h\xe2\xee\xa7\xbd\xc9o'
    b'\x1b\xf4\xaa\xc5c\xae\xba\x9f\x97\x84\xdfA\xa2;\xda[\xe4\xef\xf8\xcb'
    b'\xf1\xbd\x7f\xe1\xaf\xfa?\xe5\t\xec\xf4\xbbf_\xff\xaa)\xd9\x7f\xc9'
    b'l\xe7\x86\xe6\xac\x97\xb9\xe4\xc6\xf4\x93#\x8c_\xdd\x8f9U \x7f'
    b'\x95O\xfc9\xf8\xffo\xd2k\x03\xe8\x9f\xbc\x83\x98fm\xb1\xd5\x13'
    b'\xffv\x17\xe6\xb1\xfe]\x8a\xe4\x9fG\xbf\xb3\xfa\xbf\xfe\x1d\x1d\xf3\x12'
    b'\x8f\xfe\\\xcf\xc1\xfa\xf9\x18\xc3\xbd\xcf\xcf\x1f\x919\xa0\x01\xfd\x9a\x01'
    b'K1,\xde\xbc\xd9{\xaa\xac\x00\x00\x00'
)


def test_guess_mimetype_webp(tmp_path):
    document = new_document('<source>')
    document.settings.env = SimpleNamespace(app=SimpleNamespace(srcdir=tmp_path))
    converter = ImageConverter(document)

    file_webp = 'webp-image.webp'
    image = nodes.image(uri=file_webp, candidates={'*': file_webp})
    assert converter.guess_mimetypes(image) == ['image/webp']

    file_dat = 'webp-image.dat'
    tmp_path.joinpath(file_dat).write_bytes(WEBP_DATA)
    image = nodes.image(uri=file_dat, candidates={'*': file_dat})
    assert converter.guess_mimetypes(image) == ['image/webp']
